#!/usr/bin/env python3

from .builder import (
    BACKBONES,
    NECKS,
    HEADS,
    SEG_LOSSES,
    EDGE_LOSSES,
    JOINT_LOSSES,
    SEGMENTORS,
    build_backbone,
    build_head,
    build_neck,
    build_seg_loss,
    build_edge_loss,
    build_joint_loss,
    build_segmentor,
)

# update registries
from .backbones import *  # noqa: F401,F403
from .necks import *  # noqa: F401,F403
from .decode_heads import *  # noqa: F401,F403
from .losses import *  # noqa: F401,F403
from .segmentors import *  # noqa: F401,F403

__all__ = [
    "BACKBONES",
    "NECKS",
    "HEADS",
    "SEG_LOSSES",
    "EDGE_LOSSES",
    "JOINT_LOSSES",
    "SEGMENTORS",
    "build_backbone",
    "build_neck",
    "build_head",
    "build_seg_loss",
    "build_edge_loss",
    "build_joint_loss",
    "build_segmentor",
]



#!/usr/bin/env python3

from .builder import DATASETS
from .isprs import JointISPRSDataset, OTFJointISPRSDataset


@DATASETS.register_module()
class JointPotsdamDataset(JointISPRSDataset):
    """Joint ISPRS Potsdam dataset.
    In segmentation map annotation for Potsdam dataset, 0 is the ignore index.
    ``reduce_zero_label`` should be set to True. The ``img_suffix`` and
    ``seg_map_suffix`` are both fixed to '.png'.
    """


@DATASETS.register_module()
class OTFJointPotsdamDataset(OTFJointISPRSDataset):
    """OTF Joint ISPRS Potsdam dataset.
    In segmentation map annotation for Potsdam dataset, 0 is the ignore index.
    ``reduce_zero_label`` should be set to True. The ``img_suffix`` and
    ``seg_map_suffix`` are both fixed to '.png'.
    """



#!/usr/bin/env python3

import os
import platform

import cv2
import torch.multiprocessing as mp

from ..utils import get_root_logger


def setup_multi_processes(cfg):
    """Setup multi-processing environment variables."""
    logger = get_root_logger()

    # set multi-process start method
    if platform.system() != "Windows":
        mp_start_method = cfg.get("mp_start_method", None)
        current_method = mp.get_start_method(allow_none=True)
        if mp_start_method in ("fork", "spawn", "forkserver"):
            logger.info(
                f"Multi-processing start method `{mp_start_method}` is "
                f"different from the previous setting `{current_method}`."
                f"It will be force set to `{mp_start_method}`."
            )
            mp.set_start_method(mp_start_method, force=True)
        else:
            logger.info(f"Multi-processing start method is `{mp_start_method}`")

    # disable opencv multithreading to avoid system being overloaded
    opencv_num_threads = cfg.get("opencv_num_threads", None)
    if isinstance(opencv_num_threads, int):
        logger.info(f"OpenCV num_threads is `{opencv_num_threads}`")
        cv2.setNumThreads(opencv_num_threads)
    else:
        logger.info(f"OpenCV num_threads is `{cv2.getNumThreads}")

    if cfg.data.workers_per_gpu > 1:
        # setup OMP threads
        # This code is referred from https://github.com/pytorch/pytorch/blob/master/torch/distributed/run.py  # noqa
        omp_num_threads = cfg.get("omp_num_threads", None)
        if "OMP_NUM_THREADS" not in os.environ:
            if isinstance(omp_num_threads, int):
                logger.info(f"OMP num threads is {omp_num_threads}")
                os.environ["OMP_NUM_THREADS"] = str(omp_num_threads)
        else:
            logger.info(f'OMP num threads is {os.environ["OMP_NUM_THREADS"] }')

        # setup MKL threads
        if "MKL_NUM_THREADS" not in os.environ:
            mkl_num_threads = cfg.get("mkl_num_threads", None)
            if isinstance(mkl_num_threads, int):
                logger.info(f"MKL num threads is {mkl_num_threads}")
                os.environ["MKL_NUM_THREADS"] = str(mkl_num_threads)
        else:
            logger.info(f'MKL num threads is {os.environ["MKL_NUM_THREADS"]}')



#!/usr/bin/env python3

"""Custom EncoderDecoder that subclasses the original version.

- Support methods that depends on passing input image to the head.
- Need to use mmseg dataset class.
- Only supports mmseg registered LOSSES (SEG_LOSSES).
"""

from mmseg.models.segmentors import EncoderDecoder

from ..builder import SEGMENTORS


@SEGMENTORS.register_module()
class CustomEncoderDecoder(EncoderDecoder):
    """Custom Encoder Decoder segmentors.

    - pass input image to decode head
    """

    def __init__(
        self,
        *args,
        pass_input_image=False,
        **kwargs,
    ):
        super(CustomEncoderDecoder, self).__init__(*args, **kwargs)

        # pass input image to decode head
        self.pass_input_image = pass_input_image

    def extract_feat(self, img):
        """Extract features from images."""
        x = self.backbone(img)
        if self.with_neck:
            x = self.neck(x)

        if self.pass_input_image:
            # HACK: append input
            x = (*x, img)

        return x



#!/usr/bin/env python3

import torch


class UnNormalize(object):
    def __init__(
        self,
        mean=[123.675, 116.28, 103.53],
        std=[58.395, 57.12, 57.375],
        inplace=True,
    ):
        self.mean = mean
        self.std = std
        self.inplace = inplace

    def __call__(self, tensor):
        """
        Args:
            tensor (Tensor): Tensor image of size (C, H, W) to be normalized.
                Supports batched tensor.
        Returns:
            Tensor: Normalized image.
        """

        if tensor.ndim == 4:
            tensor = torch.permute(tensor, (1, 0, 2, 3)).contiguous()
        assert tensor.shape[0] == len(self.mean) == len(self.std)

        if self.inplace:
            for t, m, s in zip(tensor, self.mean, self.std):
                t.mul_(s).add_(m)
                # The normalize code -> t.sub_(m).div_(s)

            if tensor.ndim == 4:
                tensor = torch.permute(tensor, (1, 0, 2, 3)).contiguous()

            return tensor
        else:
            dtype = tensor.dtype
            device = tensor.device
            mean = torch.as_tensor(self.mean, dtype=dtype, device=device)
            std = torch.as_tensor(self.std, dtype=dtype, device=device)

            out = torch.zeros_like(tensor)
            for i, (t, m, s) in enumerate(zip(tensor, mean, std)):
                out[i] = t * s + m

            if out.ndim == 4:
                out = torch.permute(out, (1, 0, 2, 3)).contiguous()

            return out



#!/usr/bin/env python3

import torch
import torch.nn as nn
import torch.nn.functional as F

import numpy as np

from ..builder import SEG_LOSSES
from potato.ops.image_processing import compute_grad_mag


def perturbate_input_(input, n_elements=200):
    N, C, H, W = input.shape
    assert N == 1
    c_ = np.random.random_integers(0, C - 1, n_elements)
    h_ = np.random.random_integers(0, H - 1, n_elements)
    w_ = np.random.random_integers(0, W - 1, n_elements)
    for c_idx in c_:
        for h_idx in h_:
            for w_idx in w_:
                input[0, c_idx, h_idx, w_idx] = 1
    return input


def _sample_gumbel(shape, eps=1e-10, device=torch.device("cpu")):
    """
    Sample from Gumbel(0, 1)
    based on
    https://github.com/ericjang/gumbel-softmax/blob/3c8584924603869e90ca74ac20a6a03d99a91ef9/Categorical%20VAE.ipynb ,
    (MIT license)
    """
    U = torch.rand(shape).to(device)
    return -torch.log(eps - torch.log(U + eps))


def _gumbel_softmax_sample(logits, tau=1, eps=1e-10):
    """
    Draw a sample from the Gumbel-Softmax distribution
    based on
    https://github.com/ericjang/gumbel-softmax/blob/3c8584924603869e90ca74ac20a6a03d99a91ef9/Categorical%20VAE.ipynb
    (MIT license)
    """
    assert logits.dim() == 3
    gumbel_noise = _sample_gumbel(logits.size(), eps=eps, device=logits.device)
    y = logits + gumbel_noise
    return F.softmax(y / tau, 1)


def _one_hot_embedding(labels, num_classes):
    """Embedding labels to one-hot form.
    Args:
        labels: (LongTensor) class labels, sized [N,].
        num_classes: (int) number of classes.
    Returns:
        (tensor) encoded labels, sized [N, #classes].
    """

    device = labels.device

    y = torch.eye(num_classes).to(device)
    return y[labels].permute(0, 3, 1, 2)


@SEG_LOSSES.register_module()
class DualTaskLoss(nn.Module):
    def __init__(
        self,
        num_classes,
        loss_weight=1.0,
        loss_name="loss_dual_task",
    ):
        super(DualTaskLoss, self).__init__()
        self.num_classes = num_classes
        self.loss_weight = loss_weight
        self._loss_name = loss_name

    def forward(
        self,
        cls_score,
        label,
        weight=None,
        ignore_pixel=255,
        **kwargs,
    ):
        """
        Args:
            input_logits: NxCxHxW
            gt_semantic_masks: NxCxHxW

        Returns:
            final loss
        """
        N, C, H, W = cls_score.shape
        th = 1e-8  # 1e-10
        eps = 1e-10
        device = cls_score.device
        ignore_mask = (label == ignore_pixel).detach()
        cls_score = torch.where(
            ignore_mask.view(N, 1, H, W).expand(N, self.num_classes, H, W),
            torch.zeros(N, C, H, W).to(device),
            cls_score,
        )
        gt_semantic_masks = label.detach()
        gt_semantic_masks = torch.where(
            ignore_mask, torch.zeros(N, H, W).long().to(device), gt_semantic_masks
        )
        gt_semantic_masks = _one_hot_embedding(
            gt_semantic_masks, self.num_classes
        ).detach()

        g = _gumbel_softmax_sample(cls_score.view(N, C, -1), tau=0.5)
        g = g.reshape((N, C, H, W))
        g = compute_grad_mag(g)  # FIXME: what is this for?

        g_hat = compute_grad_mag(gt_semantic_masks)

        # general pixel-wise loss
        g = g.view(N, -1)
        g_hat = g_hat.view(N, -1)
        loss_ewise = F.l1_loss(g, g_hat, reduction="none")

        p_plus_g_mask = (g >= th).detach().float()
        loss_p_plus_g = torch.sum(loss_ewise * p_plus_g_mask) / (
            torch.sum(p_plus_g_mask) + eps
        )

        p_plus_g_hat_mask = (g_hat >= th).detach().float()
        loss_p_plus_g_hat = torch.sum(loss_ewise * p_plus_g_hat_mask) / (
            torch.sum(p_plus_g_hat_mask) + eps
        )

        total_loss = 0.5 * loss_p_plus_g + 0.5 * loss_p_plus_g_hat

        return self.loss_weight * total_loss

    @property
    def loss_name(self):
        return self._loss_name



#!/usr/bin/env python3

import random
import warnings

import numpy as np
import torch
import torch.distributed as dist

import mmcv
from mmcv.parallel import MMDataParallel, MMDistributedDataParallel
from mmcv.runner import HOOKS, build_optimizer, build_runner, get_dist_info
from mmcv.utils import build_from_cfg

from blette.apis import PatchLogBuffer

from potato import digit_version
from potato.core import (
    DistEvalHook,
    EvalHook,
    DistEvalSegHook,
    EvalSegHook,
    DistEvalEdgeHook,
    EvalEdgeHook,
)
from potato.datasets import build_dataloader, build_dataset
from potato.utils import find_latest_checkpoint, get_root_logger


def init_random_seed(seed=None, device="cuda"):
    """Initialize random seed.

    If the seed is not set, the seed will be automatically randomized,
    and then broadcast to all processes to prevent some potential bugs.
    Args:
        seed (int, Optional): The seed. Default to None.
        device (str): The device where the seed will be put on.
            Default to 'cuda'.
    Returns:
        int: Seed to be used.
    """
    if seed is not None:
        return seed

    # Make sure all ranks share the same random seed to prevent
    # some potential bugs. Please refer to
    # https://github.com/open-mmlab/mmdetection/issues/6339
    rank, world_size = get_dist_info()
    seed = np.random.randint(2**31)
    if world_size == 1:
        return seed

    if rank == 0:
        random_num = torch.tensor(seed, dtype=torch.int32, device=device)
    else:
        random_num = torch.tensor(0, dtype=torch.int32, device=device)
    dist.broadcast(random_num, src=0)
    return random_num.item()


def set_random_seed(seed, deterministic=False):
    """Set random seed.

    Args:
        seed (int): Seed to be used.
        deterministic (bool): Whether to set the deterministic option for
            CUDNN backend, i.e., set `torch.backends.cudnn.deterministic`
            to True and `torch.backends.cudnn.benchmark` to False.
            Default: False.
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    if deterministic:
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


def train_seg_det(
    model,
    dataset,
    cfg,
    distributed=False,
    validate=False,
    validate_edge=False,  # add edge validation
    timestamp=None,
    meta=None,
):
    """Launch segmentor training."""
    logger = get_root_logger(cfg.log_level)

    # prepare data loaders
    dataset = dataset if isinstance(dataset, (list, tuple)) else [dataset]
    data_loaders = [
        build_dataloader(
            ds,
            cfg.data.samples_per_gpu,
            cfg.data.workers_per_gpu,
            # cfg.gpus will be ignored if distributed
            len(cfg.gpu_ids),
            dist=distributed,
            seed=cfg.seed,
            drop_last=True,
        )
        for ds in dataset
    ]

    # put model on gpus
    if distributed:
        find_unused_parameters = cfg.get("find_unused_parameters", False)
        # Sets the `find_unused_parameters` parameter in
        # torch.nn.parallel.DistributedDataParallel
        model = MMDistributedDataParallel(
            model.cuda(),
            device_ids=[torch.cuda.current_device()],
            broadcast_buffers=False,
            find_unused_parameters=find_unused_parameters,
        )
    else:
        if not torch.cuda.is_available():
            assert digit_version(mmcv.__version__) >= digit_version(
                "1.4.4"
            ), "Please use MMCV >= 1.4.4 for CPU training!"
        model = MMDataParallel(model, device_ids=cfg.gpu_ids)
    # build runner
    optimizer = build_optimizer(model, cfg.optimizer)

    if cfg.get("runner") is None:
        cfg.runner = {"type": "IterBasedRunner", "max_iters": cfg.total_iters}
        warnings.warn(
            "config is now expected to have a `runner` section, "
            "please set `runner` in your config.",
            UserWarning,
        )

    runner = build_runner(
        cfg.runner,
        default_args=dict(
            model=model,
            batch_processor=None,
            optimizer=optimizer,
            work_dir=cfg.work_dir,
            logger=logger,
            meta=meta,
        ),
    )
    # patch log buffer for logger hooks
    runner.log_buffer = PatchLogBuffer()

    # register hooks
    runner.register_training_hooks(
        cfg.lr_config,
        cfg.optimizer_config,
        cfg.checkpoint_config,
        cfg.log_config,
        cfg.get("momentum_config", None),
    )

    # an ugly walkaround to make the .log and .log.json filenames the same
    runner.timestamp = timestamp

    # register eval hooks
    if validate:
        val_dataset = build_dataset(cfg.data.val, dict(test_mode=True))
        val_dataloader = build_dataloader(
            val_dataset,
            samples_per_gpu=1,
            workers_per_gpu=cfg.data.workers_per_gpu,
            dist=distributed,
            shuffle=False,
        )
        eval_cfg = cfg.get("evaluation", {})
        eval_cfg["by_epoch"] = cfg.runner["type"] != "IterBasedRunner"

        if cfg.model.type == "JointEncoderDecoder":
            # joint model
            eval_hook = DistEvalSegHook if distributed else EvalSegHook
        else:
            # only seg
            eval_hook = DistEvalHook if distributed else EvalHook
        # In this PR (https://github.com/open-mmlab/mmcv/pull/1193), the
        # priority of IterTimerHook has been modified from 'NORMAL' to 'LOW'.
        runner.register_hook(eval_hook(val_dataloader, **eval_cfg), priority="LOW")

    if validate_edge:
        val_dataset = build_dataset(cfg.data.val, dict(test_mode=True))

        # HACK: limit number of images for speed (edge validations are not important
        # because it's only an eye-ball estimate; for full evaluation, you would need
        # to run the edge_eval script)
        eval_img_idx = np.arange(len(val_dataset))
        np.random.shuffle(eval_img_idx)
        eval_img_idx = eval_img_idx[:50]  # NOTE: hard-coded
        img_infos = val_dataset.img_infos
        new_img_infos = []
        for i in eval_img_idx:
            new_img_infos.append(img_infos[i])
        val_dataset.img_infos = new_img_infos

        val_dataloader = build_dataloader(
            val_dataset,
            samples_per_gpu=1,
            workers_per_gpu=cfg.data.workers_per_gpu,
            dist=distributed,
            shuffle=False,
        )
        eval_cfg = cfg.get("evaluation_edge", {})  # edge eval cfg
        eval_cfg["by_epoch"] = cfg.runner["type"] != "IterBasedRunner"

        # NOTE: model needs to be either JointEncoderDecoder or BinaryJointEncoderDecoder
        # assert cfg.model.type == "JointEncoderDecoder", "needs to be joint model"
        eval_hook = DistEvalEdgeHook if distributed else EvalEdgeHook
        # In this PR (https://github.com/open-mmlab/mmcv/pull/1193), the
        # priority of IterTimerHook has been modified from 'NORMAL' to 'LOW'.
        runner.register_hook(eval_hook(val_dataloader, **eval_cfg), priority="LOW")

    # user-defined hooks
    if cfg.get("custom_hooks", None):
        custom_hooks = cfg.custom_hooks
        assert isinstance(
            custom_hooks, list
        ), f"custom_hooks expect list type, but got {type(custom_hooks)}"
        for hook_cfg in cfg.custom_hooks:
            assert isinstance(hook_cfg, dict), (
                "Each item in custom_hooks expects dict type, but got "
                f"{type(hook_cfg)}"
            )
            hook_cfg = hook_cfg.copy()
            priority = hook_cfg.pop("priority", "NORMAL")
            hook = build_from_cfg(hook_cfg, HOOKS)
            runner.register_hook(hook, priority=priority)

    if cfg.resume_from is None and cfg.get("auto_resume"):
        resume_from = find_latest_checkpoint(cfg.work_dir)
        if resume_from is not None:
            cfg.resume_from = resume_from
    if cfg.resume_from:
        runner.resume(cfg.resume_from)
    elif cfg.load_from:
        runner.load_checkpoint(cfg.load_from)
    runner.run(data_loaders, cfg.workflow)



#!/usr/bin/env python3

import os.path as osp
from copy import deepcopy

import numpy as np

import mmcv
from mmcv.runner import HOOKS
from mmcv.runner.dist_utils import master_only
from mmcv.runner.hooks.checkpoint import CheckpointHook

from mmseg.datasets.builder import build_dataloader

from blette.core.hook.basic_wandb_logger import BasicWandbHook
from potato.core.evaluation import (
    EvalHook,
    DistEvalHook,
    EvalSegHook,
    EvalEdgeHook,
    DistEvalSegHook,
    DistEvalEdgeHook,
)
from potato.visualize.vis_edge import beautify_edge


@HOOKS.register_module()
class PotatoWandbHook(BasicWandbHook):
    def __init__(
        self,
        init_kwargs=None,
        interval=50,
        log_checkpoint=False,
        log_checkpoint_metadata=False,
        num_eval_images=50,
        **kwargs,
    ):
        super().__init__(init_kwargs, interval, **kwargs)

        self.log_checkpoint = log_checkpoint
        self.log_checkpoint_metadata = log_checkpoint and log_checkpoint_metadata
        self.num_eval_images = num_eval_images
        self.log_evaluation = num_eval_images > 0
        self.ckpt_hook: CheckpointHook = None

        # NOTE: we're dealing with joint task
        self.eval_seg_hook = None
        self.test_seg_fn = None
        self.eval_edge_hook = None
        self.test_edge_fn = None

        self.num_evals = 0
        self._waited = 0

    @master_only
    def before_run(self, runner):
        super().before_run(runner)

        # Check if EvalHook(s) and CheckpointHook are available.
        # NOTE: use the single-gpu version of evaluatation function instead of multi-gpu
        for hook in runner.hooks:
            if isinstance(hook, CheckpointHook):
                self.ckpt_hook = hook

            # MMSeg EvalHooks
            if isinstance(hook, EvalHook):
                from mmseg.apis import single_gpu_test

                self.eval_seg_hook = hook
                self.test_seg_fn = single_gpu_test
            if isinstance(hook, DistEvalHook):
                from mmseg.apis import single_gpu_test

                self.eval_seg_hook = hook
                self.test_seg_fn = single_gpu_test

            # Potato EvalHooks
            if isinstance(hook, EvalSegHook):
                from potato.apis.test import single_gpu_seg_test

                self.eval_seg_hook = hook
                self.test_seg_fn = single_gpu_seg_test
            if isinstance(hook, DistEvalSegHook):
                from potato.apis.test import single_gpu_seg_test

                self.eval_seg_hook = hook
                self.test_seg_fn = single_gpu_seg_test

            # Potato edge EvalHooks
            if isinstance(hook, EvalEdgeHook):
                from potato.apis.test import single_gpu_edge_test

                self.eval_edge_hook = hook
                self.test_edge_fn = single_gpu_edge_test
            if isinstance(hook, DistEvalEdgeHook):
                from potato.apis.test import single_gpu_edge_test

                self.eval_edge_hook = hook
                self.test_edge_fn = single_gpu_edge_test

        # Check conditions to log checkpoint
        if self.log_checkpoint:
            if self.ckpt_hook is None:
                self.log_checkpoint = False
                self.log_checkpoint_metadata = False
                runner.logger.warning(
                    "To log checkpoint in PotatoWandbHook, `CheckpointHook` is"
                    "required, please check hooks in the runner."
                )
            else:
                self.ckpt_interval = self.ckpt_hook.interval

        # Check conditions to log evaluation
        if self.log_evaluation or self.log_checkpoint_metadata:
            if self.eval_seg_hook is None:
                self.log_evaluation = False
                self.log_checkpoint_metadata = False
                runner.logger.warning(
                    "To log evaluation or checkpoint metadata in "
                    "PotatoWandbHook, `EvalSegHook` or `DistEvalSegHook` in potato "
                    "is required, please check whether the validation "
                    "is enabled."
                )
            else:
                self.eval_interval = self.eval_seg_hook.interval

                # validation datasets are the same for both seg and edge eval hooks
                val_dataset = self.eval_seg_hook.dataloader.dataset

                # Determine the number of samples to be logged.
                if self.num_eval_images > len(val_dataset):
                    self.num_eval_images = len(val_dataset)
                    runner.logger.warning(
                        f"The num_eval_images ({self.num_eval_images}) is "
                        "greater than the total number of validation samples "
                        f"({len(val_dataset)}). The complete validation "
                        "dataset will be logged."
                    )

        # Check conditions to log checkpoint metadata
        if self.log_checkpoint_metadata:
            assert self.ckpt_interval % self.eval_interval == 0, (
                "To log checkpoint metadata in PotatoWandbHook, the interval "
                f"of checkpoint saving ({self.ckpt_interval}) should be "
                "divisible by the interval of evaluation "
                f"({self.eval_interval})."
            )

        if self.eval_seg_hook:
            self.num_evals += 1
        if self.eval_edge_hook:
            self.num_evals += 1

        # Initialize evaluation table
        if self.log_evaluation:
            self._initialize_gts(runner)

    @master_only
    def after_train_iter(self, runner):
        if self.get_mode(runner) == "train":
            # An ugly patch. The iter-based eval hook will call the
            # `after_train_iter` method of all logger hooks before evaluation.
            # Use this trick to skip that call.
            # Don't call super method at first, it will clear the log_buffer
            return super().after_train_iter(runner)
        else:
            super().after_train_iter(runner)

        # Even if `runner.mode` is "train", the mode could be "val"
        # depending on if "time" is in `runner.log_buffer.output`

        if self.by_epoch:
            return

        # HACK: `after_train_iter` gets called after evaluating one of the EvalHooks,
        # so if there are multiple EvalHooks, it will evaluate for every hook
        # need to make sure that it runs after the last iter
        if self._waited > self.num_evals:
            print("something wrong happened")

        self._waited += 1
        if self._waited < self.num_evals:
            # print(f"waiting for more evals (waited {self._waited})")
            return
        self._waited = 0

        # print(f'passed block {self._waited}')

        # Save checkpoint and metadata
        if self.log_checkpoint and (
            self.every_n_iters(runner, self.ckpt_interval)
            or (self.ckpt_hook.save_last and self.is_last_iter(runner))
        ):
            if self.log_checkpoint_metadata and self.eval_seg_hook:
                metadata = {"iter": runner.iter + 1, **self._get_eval_results()}
            else:
                metadata = None
            aliases = [f"iter_{runner.iter+1}", "latest"]
            model_path = osp.join(self.ckpt_hook.out_dir, f"iter_{runner.iter+1}.pth")
            self._log_ckpt_as_artifact(model_path, aliases, metadata)

        # Save seg prediction table
        if self.eval_seg_hook:
            if self.log_evaluation and self.eval_seg_hook._should_evaluate(runner):
                self._log_seg_predictions(runner)

        # Save edge prediction table
        if self.eval_edge_hook:
            if self.log_evaluation and self.eval_edge_hook._should_evaluate(runner):
                self._log_edge_predictions(runner)

    @master_only
    def after_run(self, runner):
        self.wandb.finish()

    def _log_ckpt_as_artifact(self, model_path, aliases, metadata=None):
        """Log model checkpoint as  W&B Artifact.
        Args:
            model_path (str): Path of the checkpoint to log.
            aliases (list): List of the aliases associated with this artifact.
            metadata (dict, optional): Metadata associated with this artifact.
        """
        model_artifact = self.wandb.Artifact(
            f"run_{self.wandb.run.id}_model", type="model", metadata=metadata
        )
        model_artifact.add_file(model_path)
        self.wandb.log_artifact(model_artifact, aliases=aliases)

    def _get_eval_results(self):
        """Get model evaluation results."""
        seg_results = self.eval_seg_hook.latest_results
        val_dataset = self.eval_seg_hook.dataloader.dataset
        eval_seg_results = val_dataset.evaluate(
            seg_results,
            logger="silent",
            **self.eval_seg_hook.eval_kwargs,
        )
        results = dict(seg=eval_seg_results)

        if self.eval_edge_hook:
            edge_results = self.eval_edge_hook.latest_results
            val_dataset = self.eval_edge_hook.dataloader.dataset
            eval_edge_results = val_dataset.evaluate(
                edge_results,
                logger="silent",
                **self.eval_edge_hook.eval_kwargs,
            )
            results.update(edge=eval_edge_results)

        return results

    def _initialize_gts(self, runner):
        # validation datasets are the same for both seg and edge eval hooks
        val_dataset = deepcopy(self.eval_seg_hook.dataloader.dataset)

        # Prune dataset
        eval_image_idx = np.arange(len(val_dataset))
        # Set seed so that same validation set is logged each time.
        np.random.seed(42)
        np.random.shuffle(eval_image_idx)
        eval_image_idx = eval_image_idx[: self.num_eval_images]

        img_infos = val_dataset.img_infos
        new_img_infos = []
        for i in eval_image_idx:
            new_img_infos.append(img_infos[i])
        val_dataset.img_infos = new_img_infos

        self.dataset = val_dataset

        # Setting up a dataloader
        # FIXME: not sure about the arguments being passed
        self.dataloader = build_dataloader(
            dataset=val_dataset,
            samples_per_gpu=1,
            workers_per_gpu=2,
            dist=False,  # `master_only``
            shuffle=False,
        )

        # Get image loading pipeline
        from mmseg.datasets.pipelines import LoadImageFromFile

        img_loader = None
        for t in self.dataset.pipeline.transforms:
            if isinstance(t, LoadImageFromFile):
                img_loader = t

        if img_loader is None:
            self.log_evaluation = False
            runner.logger.warning(
                "LoadImageFromFile is required to add images " "to W&B Tables."
            )
            return

        classes = self.dataset.CLASSES
        self.class_id_to_label = {id: name for id, name in enumerate(classes)}
        self.class_set = self.wandb.Classes(
            [{"id": id, "name": name} for id, name in self.class_id_to_label.items()]
        )

        # Init data table.
        if self.eval_edge_hook is not None:
            columns = ["image_name", "image", "edge"]
        else:
            columns = ["image_name", "image"]
        data_table = self.wandb.Table(columns=columns)

        for idx, img_info in enumerate(self.dataset.img_infos):
            image_name = img_info["filename"]

            # Get image and convert from BGR to RGB
            img_meta = img_loader(
                dict(img_info=img_info, img_prefix=self.dataset.img_dir)
            )
            image = mmcv.bgr2rgb(img_meta["img"])

            # Get segmentation mask
            seg_mask = self.dataset.get_gt_seg_map_by_idx(idx)
            assert (
                seg_mask.ndim == 2
            ), f"ERR: need seg_mask to be 2dim but got {seg_mask.ndim}dim"
            # Dict of masks to be logged.
            wandb_masks = {
                "ground_truth": {
                    "mask_data": seg_mask,
                    "class_labels": self.class_id_to_label,
                }
            }

            if self.eval_edge_hook is not None:
                edge_mask = self.dataset.get_gt_edge_map_by_idx(idx)
                edge = beautify_edge(
                    edge_mask,
                    palette=self.dataset.PALETTE,
                    beautify_threshold=0.7,
                )

                # Log a row to the data table.
                data_table.add_data(
                    image_name,
                    self.wandb.Image(image, masks=wandb_masks, classes=self.class_set),
                    self.wandb.Image(edge),
                )
            else:
                # Log a row to the data table.
                data_table.add_data(
                    image_name,
                    self.wandb.Image(image, masks=wandb_masks, classes=self.class_set),
                )

        """Log the W&B Tables for validation data as artifact and calls
        `use_artifact` on it so that the evaluation table can use the reference
        of already uploaded images.
        This allows the data to be uploaded just once.
        """
        data_artifact = self.wandb.Artifact("val", type="dataset")
        data_artifact.add(data_table, "val_data")
        self.wandb.run.use_artifact(data_artifact)
        data_artifact.wait()  # might take a while...
        self.data_table_ref = data_artifact.get("val_data")

    def _log_seg_predictions(self, runner):
        results = self.test_seg_fn(runner.model, self.dataloader)
        columns = ["image_name", "ground_truth", "prediction"]
        eval_table = self.wandb.Table(columns=columns)
        table_idxs = self.data_table_ref.get_index()
        assert len(table_idxs) == len(self.dataset)
        assert len(results) == len(self.dataset)

        for ndx, pred_mask in enumerate(results):
            assert pred_mask.ndim == 2, f"{pred_mask.ndim}dim is not supported"
            wandb_masks = {
                "prediction": {
                    "mask_data": pred_mask,
                    "class_labels": self.class_id_to_label,
                }
            }

            # Log a row to the data table.
            eval_table.add_data(
                self.data_table_ref.data[ndx][0],
                self.data_table_ref.data[ndx][1],
                self.wandb.Image(
                    self.data_table_ref.data[ndx][1],
                    masks=wandb_masks,
                    classes=self.class_set,
                ),
            )

        pred_artifact = self.wandb.Artifact(
            f"run_{self.wandb.run.id}_pred", type="evaluation"
        )
        pred_artifact.add(eval_table, "eval_data")  # NOTE: naming
        self.wandb.run.log_artifact(pred_artifact)

    def _log_edge_predictions(self, runner):
        results = self.test_edge_fn(runner.model, self.dataloader)
        columns = ["image_name", "ground_truth", "prediction"]
        eval_table = self.wandb.Table(columns=columns)
        table_idxs = self.data_table_ref.get_index()
        assert len(table_idxs) == len(self.dataset)
        assert len(results) == len(self.dataset)

        for ndx, pred_mask in enumerate(results):
            # Convert pred_mask
            if pred_mask.ndim == 3:
                if pred_mask.shape[0] == 1:
                    pred_mask = pred_mask.squeeze(0)
                else:
                    pred_mask = beautify_edge(
                        edges=pred_mask,
                        palette=self.dataset.PALETTE,
                        beautify_threshold=0.7,  # HARD-Coded
                    )
            # Log a row to the data table.
            eval_table.add_data(
                self.data_table_ref.data[ndx][0],
                # self.data_table_ref.data[ndx][1],  # img
                self.data_table_ref.data[ndx][2],  # edge gt
                self.wandb.Image(pred_mask),
            )
        pred_artifact = self.wandb.Artifact(
            f"run_{self.wandb.run.id}_edge_pred", type="evaluation"
        )
        pred_artifact.add(eval_table, "eval_edge_data")
        self.wandb.run.log_artifact(pred_artifact)



#!/usr/bin/env python3

"""builder.py

Currently a placeholder for subclassing `mmseg` and `blette`.
"""

# I want to use the original segmentation only datasets
from mmseg.datasets import (
    DATASETS,
    build_dataloader,
    build_dataset,
)

# blette pipeline includes mmseg's pipelines
from blette.datasets import PIPELINES


__all__ = [
    "DATASETS",
    "PIPELINES",
    "build_dataloader",
    "build_dataset",
]



#!/usr/bin/env python3

from mmseg.models.losses.accuracy import Accuracy, accuracy
from blette.models.losses.accuracy import calc_metrics as edge_accuracy

__all__ = ["Accuracy", "accuracy", "edge_accuracy"]



#!/usr/bin/env python3

import torch
import torch.nn.functional as F


def maskgrad(
    seg_map: torch.Tensor,
    sensitivity: int = 100,
) -> torch.Tensor:
    """Calculate Mask Gradients."""
    # convert to probability
    seg_probs = F.softmax(sensitivity * seg_map, dim=1)
    # average or maxpooling?
    seg_pool = F.avg_pool2d(seg_probs, kernel_size=3, stride=1, padding=1)
    mask_edge = torch.abs(seg_probs - seg_pool)  # (b, class, h, w)
    return mask_edge



#!/usr/bin/env python3

import torch
import torch.nn as nn

from mmcv.cnn import ConvModule
from mmcv.runner import force_fp32
from mmseg.ops import resize

from potato.models.builder import HEADS, build_edge_loss
from potato.models.utils import maskgrad
from .manyedge_decode_head import BaseManyEdgeDecodeHead


class PatchContext(nn.Module):
    """PatchContext Module

    Input image ->
        make pyramid patches (avgpool)s ->
        conv ->
        sum ->
        element wise sum on original feature


    Original Paper:
    - I don't think they added any bn or activations
    """

    def __init__(self, channels, patch_sizes=(1, 3, 5, 7), **kwargs):
        super().__init__()
        assert isinstance(patch_sizes, (list, tuple))
        assert len(patch_sizes) > 0

        self.patch_sizes = patch_sizes

        modules = []
        for patch in patch_sizes:
            if isinstance(patch, int):
                patch = (patch, patch)
            assert isinstance(patch, (list, tuple))
            assert len(patch) == 2
            modules.append(
                nn.Sequential(
                    nn.AdaptiveAvgPool2d(patch),
                    ConvModule(
                        in_channels=channels,
                        out_channels=channels,
                        kernel_size=patch,
                        **kwargs,
                    ),
                )
            )
        self.patch_convs = nn.ModuleList(modules)

    def forward(self, x):
        b, c, h, w = x.shape
        ps = []
        for pconv in self.patch_convs:
            xi = pconv(x)  # (b, c, 1, 1)
            ps.append(xi)

        ps = torch.stack(ps, dim=0).sum(dim=0)
        return x + ps.repeat(1, 1, h, w)


class PCM(nn.Module):
    """Pyramid Context Module

    Seems like bn+relu in PatchContext stabilizes the training
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        num_inputs,
        patch_sizes=(1, 3, 5, 7),
        conv_cfg=None,
        norm_cfg=None,
        act_cfg=dict(type="ReLU"),
    ):
        super().__init__()

        # lowest -> 4 inputs
        # second -> 3 inputs
        # top -> 2 inputs

        # levels
        # res4 -> l0
        # res2 -> l1
        # res1 -> l2

        # each level will have 3x3 conv (bn + relu)
        self.convs = torch.nn.ModuleList(
            [
                ConvModule(
                    in_channels=in_channels,
                    out_channels=out_channels,
                    kernel_size=3,
                    padding=1,
                    conv_cfg=conv_cfg,
                    norm_cfg=norm_cfg,
                    act_cfg=act_cfg,
                )
                for _ in range(num_inputs)
            ]
        )

        if num_inputs > 1:
            pc_kwargs = dict(
                channels=out_channels,
                patch_sizes=patch_sizes,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
                bias=False,
                act_cfg=act_cfg,
            )

            # only l0 inputs
            # apply PatchContext to `l0 s-2
            patch_contexts = [PatchContext(**pc_kwargs)]

            if num_inputs == 3:
                # l0 and l1
                # apply PatchContext to `l0 s-1` and `l1 s-2`
                patch_contexts.append(PatchContext(**pc_kwargs))
            elif num_inputs == 4:
                # l0, l1, and l2
                # apply PatchContext to `l0 s-1`, `l1 s-1`, and `l2 s-2`
                patch_contexts.extend(
                    [
                        PatchContext(**pc_kwargs),
                        PatchContext(**pc_kwargs),
                    ]
                )
            self.patch_contexts = nn.ModuleList(patch_contexts)
        else:
            self.patch_contexts = None

    def forward(self, x):
        assert isinstance(x, (list, tuple))

        xs = []
        for xi, conv in zip(x, self.convs):
            xi = conv(xi)
            xs.append(xi)

        if len(xs) == 1:
            return xs[0]

        # assume that `s-2` is the top
        s2 = xs.pop(0)
        _, _, H, W = s2.shape

        outs = []
        for xi, pconv in zip(xs, self.patch_contexts):
            xi = pconv(xi)

            _, _, h, w = xi.shape
            if h < H or w < W:
                xi = resize(xi, (H, W), mode="bilinear", align_corners=False)

            xi = s2 * xi
            outs.append(xi)

        outs = torch.stack(outs, dim=0).sum(dim=0)
        s2 = s2 + outs
        return s2


class Step(nn.Module):
    def __init__(
        self,
        channels,
        patch_sizes,
        conv_cfg=None,
        norm_cfg=None,
        act_cfg=dict(type="ReLU"),
    ):
        super().__init__()
        modules = [
            PCM(
                in_channels=channels,
                out_channels=channels,
                num_inputs=i + 2,
                patch_sizes=patch_sizes,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
                act_cfg=act_cfg,
            )
            for i in range(3)
        ]
        self.pcms = nn.ModuleList(modules)

    def forward(self, s2, s1):
        # l2 -> l0
        outs = []
        for i, l_s2 in enumerate(reversed(s2)):
            l = 3 - (i + 1)
            inputs = (l_s2, *s1[l:])  # `s-2` is always the first
            outs.append(self.pcms[i](inputs))

        return tuple(reversed(outs))


@HEADS.register_module()
class RPCHead(BaseManyEdgeDecodeHead):
    def __init__(
        self,
        num_steps=4,
        patch_sizes=[1, 3, 5, 7],
        edge_key="edge",
        log_edge_keys=["edge"],
        loss_duality=dict(type="DualityLoss", loss_weight=1.0),
        **kwargs,
    ):
        super().__init__(
            input_transform="multiple_select",
            edge_key=edge_key,
            log_edge_keys=log_edge_keys,
            multilabel_edge_keys=["edge"],
            no_conv_seg=True,
            **kwargs,
        )

        assert num_steps > 1

        self.side1 = ConvModule(
            in_channels=self.in_channels[0],
            out_channels=self.channels,
            kernel_size=3,
            padding=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )
        self.side2 = ConvModule(
            in_channels=self.in_channels[1],
            out_channels=self.channels,
            kernel_size=3,
            padding=1,
            conv_cfg=self.conv_cfg,
            act_cfg=self.act_cfg,
        )
        self.side3 = ConvModule(
            in_channels=self.in_channels[2],
            out_channels=self.channels,
            kernel_size=3,
            padding=1,
            conv_cfg=self.conv_cfg,
            act_cfg=self.act_cfg,
        )

        self.s1_l0 = PCM(
            in_channels=self.channels,
            out_channels=self.channels,
            num_inputs=1,
            patch_sizes=patch_sizes,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )
        self.s1_l1 = PCM(
            in_channels=self.channels,
            out_channels=self.channels,
            num_inputs=2,
            patch_sizes=patch_sizes,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )
        self.s1_l2 = PCM(
            in_channels=self.channels,
            out_channels=self.channels,
            num_inputs=3,
            patch_sizes=patch_sizes,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )

        if num_steps - 2 > 0:
            self.steps = nn.ModuleList(
                [
                    Step(
                        channels=self.channels,
                        patch_sizes=patch_sizes,
                        conv_cfg=self.conv_cfg,
                        norm_cfg=self.norm_cfg,
                        act_cfg=self.act_cfg,
                    )
                    for _ in range(num_steps - 2)
                ]
            )
        else:
            self.steps = None

        self.last_step = PCM(
            in_channels=self.channels,
            out_channels=self.channels,
            num_inputs=4,
            patch_sizes=patch_sizes,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )

        # self.prob_conv = nn.ModuleList(
        #     nn.Conv2d(self.channels, self.num_classes, kernel_size=1) for _ in range(num_steps)
        # )
        self.edge_prob = nn.Conv2d(self.channels, self.num_classes, kernel_size=1)
        self.seg_prob = nn.Conv2d(self.channels, self.num_classes, kernel_size=1)
        self.grouped_edge_conv = nn.Conv2d(
            self.num_classes * 2,
            self.num_classes,
            kernel_size=1,
            groups=self.num_classes,
        )

        self.dual_loss = build_edge_loss(loss_duality)

    def forward(self, inputs):
        """Forward function"""

        # [layer1, layer2, layer3, layer4]: 1/4, 1/8, 1/16, 1/16
        x1, x2, x3, x4 = [i for i in inputs]

        b1 = self.side1(x1)
        b2 = self.side2(x2)
        b3 = self.side3(x4)

        # run first step
        s2 = (b1, b2, b3)
        s1_l0 = self.s1_l0((b3,))
        s1_l1 = self.s1_l1((b2, b3))
        s1_l2 = self.s1_l2((b1, b2, b3))
        s1 = (s1_l2, s1_l1, s1_l0)
        outs = [s1_l2]

        if self.steps is not None:
            for step in self.steps:
                # unless the tensors of s2 are changed inplace, it should be fine
                b1, b2, b3 = s1
                s1 = step(s2, s1)
                s2 = (b1, b2, b3)
                outs.append(s1[0])

        final = self.last_step((s2[0], *s1))
        outs.append(final)

        segs = []
        edges = []
        # for i, (out, fconv) in enumerate(zip(outs, self.prob_conv)):
        #     if i % 2 == 0:
        #         # even output is for segmentation
        #         segs.append(fconv(out))
        #     else:
        #         # odd output is for edge
        #         edges.append(fconv(out))
        for i, out in enumerate(outs):
            if i % 2 == 0:
                segs.append(out)
            else:
                edges.append(out)

        # TODO: do deep supervision later
        seg = self.seg_prob(segs[-1])
        edge = self.edge_prob(edges[-1])

        mask_edge = maskgrad(
            seg_map=seg,
            sensitivity=100,
        )
        # edge = torch.relu(edge)

        # sliced concatenation and k-grouped convolution
        hold = []
        for c in range(edge.size(1)):
            hold.append(edge[:, c, ...].unsqueeze(1))
            hold.append(mask_edge[:, c, ...].unsqueeze(1))
        edge = torch.cat(hold, dim=1)
        edge = self.grouped_edge_conv(edge)  # logits

        return (seg, dict(edge=edge), mask_edge)

    def forward_train(
        self, inputs, img_metas, gt_semantic_seg, gt_semantic_edge, train_cfg
    ):
        seg_logits, semantic_edge_logits, mask_edge = self(inputs)

        losses = self.seg_losses(
            seg_logit=seg_logits,
            seg_label=gt_semantic_seg,
        )
        losses.update(
            self.multilabel_edge_losses(
                edge_logit=semantic_edge_logits,
                edge_label=gt_semantic_edge,
            )
        )
        losses.update(
            self.compute_dual_loss(
                edge_logit=semantic_edge_logits,
                maskgrad_prob=mask_edge,
            )
        )
        return losses

    def forward_test(self, inputs, img_metas, test_cfg, return_edge=False, **kwargs):
        seg, semantic_edge, mask_edge = self(inputs)
        if return_edge:
            edge = None
            if isinstance(semantic_edge, dict):
                edge = semantic_edge.get(self.edge_key, None)
            assert edge is not None, f"could not find {self.edge_key}"
            return dict(seg=seg, edge=edge)
        else:
            return dict(seg=seg)

    @force_fp32(apply_to=("edge_logit", "maskgrad_prob"))
    def compute_dual_loss(self, edge_logit, maskgrad_prob):
        edge_logit = edge_logit["edge"]
        return {
            self.dual_loss.loss_name: self.dual_loss(
                edge_logit=edge_logit,
                maskgrad=maskgrad_prob,
            )
        }



#!/usr/bin/env python3

"""Formatting

- use `DefaultFormatBundle` for segmentation only tasks
- use `JointFormatBundle` for joint segmentation and edge detection
"""

import numpy as np
from mmcv.parallel import DataContainer as DC
from mmseg.datasets.pipelines.formatting import to_tensor
from blette.datasets.pipelines import FormatImage

from potato.datasets.builder import PIPELINES


@PIPELINES.register_module()
class FormatJoint(object):
    """Format multilabel edge and segmentation"""

    def __call__(self, results):
        # non-otf
        if "gt_semantic_edge" in results:
            edge = results["gt_semantic_edge"]

            # HACK: decode RGB to 24bit array
            # it's only possible to encode 24 classes
            edge = np.unpackbits(
                edge,
                axis=2,
            )[:, :, -1 : -(results["num_classes"] + 1) : -1]
            edge = np.ascontiguousarray(edge.transpose(2, 0, 1))

            # convert to long
            results["gt_semantic_edge"] = DC(
                to_tensor(edge.astype(np.int64)),
                stack=True,
            )

            # convert segmentation mask as well
            if "gt_semantic_seg" in results:
                # convert to long
                results["gt_semantic_seg"] = DC(
                    to_tensor(results["gt_semantic_seg"][None, ...].astype(np.int64)),
                    stack=True,
                )

        elif "gt_semantic_seg" in results:
            mask2edge = results.get("mask2edge", None)
            assert mask2edge, "ERR: no mask2edge inside `results`"

            if results["inst_sensitive"]:
                inst_map = results.get("gt_inst_seg", None)
                assert inst_map is not None, "ERR: instance map is not available"
                out = mask2edge(
                    mask=results["gt_semantic_seg"],
                    inst_mask=inst_map,
                )
                # remove it from memory?
                del results["gt_inst_seg"]
            else:
                out = mask2edge(mask=results["gt_semantic_seg"])

            # out is a dict('mask'=..., 'edge'=...)
            results["gt_semantic_seg"] = DC(
                to_tensor(out["mask"][None, ...].astype(np.int64)),
                stack=True,
            )
            results["gt_semantic_edge"] = DC(
                to_tensor(out["edge"].astype(np.int64)),
                stack=True,
            )

        return results


@PIPELINES.register_module()
class FormatJointBinaryEdge:
    """Format binary edge and segmentation"""

    def __init__(
        self,
        selected_label=None,
    ):
        """
        Args:
            select_label Optional(int): if multilabel edges are given,
                the label set with this argument will be used
                (assuming that the edge is indexed from 0)
        """
        self._label = selected_label
        if self._label is not None:
            assert isinstance(self._label, int)

    def __call__(self, results):
        # non-otf
        if "gt_semantic_edge" in results:
            edge = results["gt_semantic_edge"]

            if edge.ndim == 3:
                assert self._label is not None
                # ordered (h, w, trainId)
                edge = edge[:, :, self._label]

            assert edge.ndim == 2
            # No need for hacks to convert dataset
            results["gt_semantic_edge"] = DC(
                to_tensor(edge[None, ...].astype(np.int64)),
                stack=True,
            )

            # convert segmentation mask as well
            if "gt_semantic_seg" in results:
                # convert to long
                results["gt_semantic_seg"] = DC(
                    to_tensor(results["gt_semantic_seg"][None, ...].astype(np.int64)),
                    stack=True,
                )
        elif "gt_semantic_seg" in results:
            mask2edge = results.get("mask2edge", None)
            assert mask2edge, "ERR: no mask2edge inside `results`"

            if results.get("inst_sensitive", False):
                inst_map = results.get("gt_inst_seg", None)
                assert inst_map is not None, "ERR: instance map is not available"
                out = mask2edge(
                    mask=results["gt_semantic_seg"],
                    inst_mask=inst_map,
                )
                # remove it from memory?
                del results["gt_inst_seg"]
            else:
                out = mask2edge(mask=results["gt_semantic_seg"])

            edge = out["edge"]

            if edge.ndim == 3:
                assert self._label is not None
                # ordered (trainId, h, w)
                edge = edge[self._label]

            assert edge.ndim == 2

            results["gt_semantic_edge"] = DC(
                to_tensor(edge[None, ...].astype(np.int64)),
                stack=True,
            )
            results["gt_semantic_seg"] = DC(
                to_tensor(out["mask"][None, ...].astype(np.int64)),
                stack=True,
            )

        return results


@PIPELINES.register_module()
class JointFormatBundle(object):
    def __init__(self):
        self.format_image = FormatImage()
        self.format_joint = FormatJoint()

    def __call__(self, results):
        results = self.format_image(results)
        results = self.format_joint(results)
        return results

    def __repr__(self):
        return self.__class__.__name__


@PIPELINES.register_module()
class JointBinaryFormatBundle(object):
    def __init__(self, selected_label=None):
        self.format_image = FormatImage()
        self.format_joint = FormatJointBinaryEdge(selected_label=selected_label)

    def __call__(self, results):
        results = self.format_image(results)
        results = self.format_joint(results)
        return results

    def __repr__(self):
        return self.__class__.__name__



#!/usr/bin/env python3

from .mmseg_eval_hooks import DistEvalHook, EvalHook
from .seg_eval_hooks import DistEvalSegHook, EvalSegHook
from .edge_eval_hooks import DistEvalEdgeHook, EvalEdgeHook
from .f_boundary import (
    eval_mask_boundary,
    eval_mask_boundary_batch,
    pre_eval_to_boundary,
)

__all__ = [
    "EvalHook",
    "DistEvalHook",
    "EvalSegHook",
    "DistEvalSegHook",
    "EvalEdgeHook",
    "DistEvalEdgeHook",
    "eval_mask_boundary",
    "eval_mask_boundary_batch",
    "pre_eval_to_boundary",
]



#!/usr/bin/env python3

"""Shape Stream (GSCNN)

Differences:
- Added activations between blocks
- Separate edge detection and features used in later layers
"""

import torch.nn as nn

from mmcv.cnn import ConvModule, build_activation_layer
from mmseg.ops import resize
from mmseg.models.backbones.resnet import BasicBlock

from .gated_conv import GatedSpatialConv2d


class ResNetGSCNNShapeStream(nn.Module):
    """GSCNN ShapeStream for ResNet Backbone

    Original implementation uses WiderResNet with features:
    (64, 256, 512, 4096)

    For ResNet, we use (stem, res1, res2, res4) which is:
    (64, 256, 512, 2048)

    Returns:
        binary feat, edge pred.
    """

    def __init__(
        self,
        in_channels,  # (64, 256, 512, 1024, 2048)
        num_classes,
        conv_cfg=None,
        norm_cfg=None,
        act_cfg=dict(type="ReLU"),
        interpolation="bilinear",
        align_corners=False,
    ):
        super().__init__()

        assert isinstance(in_channels, (tuple, list))
        assert len(in_channels) >= 4

        self._interp = interpolation
        self._align_corners = align_corners

        self.dsn1 = ConvModule(
            in_channels=in_channels[0],  # 64
            out_channels=64,  # no feature reduction
            kernel_size=1,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,  # bias=True in original
            act_cfg=act_cfg,
        )
        self.dsn2 = ConvModule(
            in_channels=in_channels[1],  # 256
            out_channels=1,  # gating feature
            kernel_size=1,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,  # bias=True in original
            act_cfg=act_cfg,
        )
        self.dsn3 = ConvModule(
            in_channels=in_channels[2],  # 512
            out_channels=1,  # gating feature
            kernel_size=1,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,  # bias=True in original
            act_cfg=act_cfg,
        )
        self.dsn4 = ConvModule(
            in_channels=in_channels[4],  # 2048
            out_channels=1,  # gating feature
            kernel_size=1,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,  # bias=True in original
            act_cfg=act_cfg,
        )
        self.res1 = BasicBlock(
            inplanes=in_channels[0],  # 64
            planes=in_channels[0],  # 64
            stride=1,
            downsample=None,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
        )
        self.d1 = ConvModule(
            in_channels=in_channels[0],
            out_channels=32,
            kernel_size=1,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg,
        )
        self.res2 = BasicBlock(
            inplanes=32,
            planes=32,
            stride=1,
            downsample=None,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
        )
        self.d2 = ConvModule(
            in_channels=32,
            out_channels=16,
            kernel_size=1,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg,
        )
        self.res3 = BasicBlock(
            inplanes=16,
            planes=16,
            stride=1,
            downsample=None,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
        )
        self.d3 = ConvModule(
            in_channels=16,
            out_channels=8,
            kernel_size=1,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg,
        )
        self.gate1 = GatedSpatialConv2d(
            in_channels=32,
            out_channels=32,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg,
        )
        self.gate2 = GatedSpatialConv2d(
            in_channels=16,
            out_channels=16,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg,
        )
        self.gate3 = GatedSpatialConv2d(
            in_channels=8,
            out_channels=8,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg,
        )
        self.fuse = ConvModule(
            in_channels=8,
            out_channels=num_classes,
            kernel_size=1,
            conv_cfg=conv_cfg,
            norm_cfg=None,  # norm_cfg,
            bias=True,  # originally false
            act_cfg=None,
        )

        self.act = build_activation_layer(act_cfg)

    def forward(self, x):
        bs, c, h, w = x[-1].shape
        resize_to = (h, w)  # TODO: might be too large

        # side outputs
        side1 = resize(
            self.dsn1(x[0]),
            size=resize_to,
            mode=self._interp,
            align_corners=self._align_corners,
        )
        side2 = resize(
            self.dsn2(x[1]),
            size=resize_to,
            mode=self._interp,
            align_corners=self._align_corners,
        )
        side3 = resize(
            self.dsn3(x[2]),
            size=resize_to,
            mode=self._interp,
            align_corners=self._align_corners,
        )
        side4 = resize(
            self.dsn4(x[4]),
            size=resize_to,
            mode=self._interp,
            align_corners=self._align_corners,
        )

        cs = self.d1(
            resize(
                self.res1(side1),
                size=resize_to,
                mode=self._interp,
                align_corners=self._align_corners,
            )
        )
        cs = self.gate1(cs, side2)
        cs = self.act(cs)
        cs = self.d2(
            resize(
                self.res2(cs),
                size=resize_to,
                mode=self._interp,
                align_corners=self._align_corners,
            )
        )
        cs = self.gate2(cs, side3)
        cs = self.act(cs)
        cs = self.d3(
            resize(
                self.res3(cs),
                size=resize_to,
                mode=self._interp,
                align_corners=self._align_corners,
            )
        )
        cs = self.gate3(cs, side4)
        cs = self.act(cs)
        edge_out = self.fuse(cs)

        return edge_out



#!/usr/bin/env python3

import torch.nn.functional as F
from mmseg.ops import resize

from ..builder import SEGMENTORS
from .base import BaseJointSegDet


@SEGMENTORS.register_module()
class BinaryJointEncoderDecoder(BaseJointSegDet):
    """JointEncoderDecoder for binary datasets

    - Segmentation maps have two classes (background, target)
    - Edges only have one class (edge or not; binary)
    """

    num_classes = 2
    num_edge_classes = 1

    def __init__(
        self,
        **kwargs,
    ):
        super(BinaryJointEncoderDecoder, self).__init__(**kwargs)
        assert (
            self.num_classes == 2
        ), f"for binary segmentation, num_classes should be 2, but got {self.num_classes}"

    def forward_train(
        self,
        img,
        img_metas,
        gt_semantic_seg,
        gt_semantic_edge,
        **kwargs,
    ):
        """Forward function for training.

        Args:
            img (Tensor): Input images.
            img_metas (list[dict]): List of image info dict where each dict
                has: 'img_shape', 'scale_factor', 'flip', and may also contain
                'filename', 'ori_shape', 'pad_shape', and 'img_norm_cfg'.
                For details on the values of these keys see
                `potato/datasets/pipelines/formatting.py:Collect`.
            gt_semantic_seg (Tensor): Semantic segmentation masks
                used if the architecture supports semantic segmentation task.

        Returns:
            dict[str, Tensor]: a dictionary of loss components
        """

        if len(gt_semantic_edge.shape) == 3:
            # might not have channel dim
            gt_semantic_edge = gt_semantic_edge.unsqueeze(1)

        x = self.extract_feat(img)

        losses = dict()

        loss_decode = self._decode_head_forward_train(
            x, img_metas, gt_semantic_seg, gt_semantic_edge
        )
        losses.update(loss_decode)

        if self.with_auxiliary_head:
            loss_aux = self._auxiliary_head_forward_train(x, img_metas, gt_semantic_seg)
            losses.update(loss_aux)

        return losses

    # TODO refactor
    def slide_inference(
        self,
        img,
        img_meta,
        rescale,
        return_edge=False,
        **kwargs,
    ):
        """Inference by sliding-window with overlap.

        If h_crop > h_img or w_crop > w_img, the small patch will be used to
        decode without padding.
        """

        h_stride, w_stride = self.test_cfg.stride
        h_crop, w_crop = self.test_cfg.crop_size
        batch_size, _, h_img, w_img = img.size()
        num_classes = self.num_classes

        h_grids = max(h_img - h_crop + h_stride - 1, 0) // h_stride + 1
        w_grids = max(w_img - w_crop + w_stride - 1, 0) // w_stride + 1
        seg_preds = img.new_zeros((batch_size, num_classes, h_img, w_img))
        if return_edge:
            num_edge_classes = self.num_edge_classes
            edge_preds = img.new_zeros((batch_size, num_edge_classes, h_img, w_img))

        count_mat = img.new_zeros((batch_size, 1, h_img, w_img))
        for h_idx in range(h_grids):
            for w_idx in range(w_grids):
                y1 = h_idx * h_stride
                x1 = w_idx * w_stride
                y2 = min(y1 + h_crop, h_img)
                x2 = min(x1 + w_crop, w_img)
                y1 = max(y2 - h_crop, 0)
                x1 = max(x2 - w_crop, 0)
                crop_img = img[:, :, y1:y2, x1:x2]
                logits = self.encode_decode(
                    crop_img, img_meta, return_edge=return_edge, **kwargs
                )

                if isinstance(logits, tuple):
                    seg_logit = logits[0]
                    edge = logits[1]
                    seg_preds += F.pad(
                        seg_logit,
                        (
                            int(x1),
                            int(seg_preds.shape[3] - x2),
                            int(y1),
                            int(seg_preds.shape[2] - y2),
                        ),
                    )
                    edge_preds += F.pad(
                        edge,
                        (
                            int(x1),
                            int(edge_preds.shape[3] - x2),
                            int(y1),
                            int(edge_preds.shape[2] - y2),
                        ),
                    )
                else:
                    seg_preds += F.pad(
                        logits,
                        (
                            int(x1),
                            int(seg_preds.shape[3] - x2),
                            int(y1),
                            int(seg_preds.shape[2] - y2),
                        ),
                    )

                count_mat[:, :, y1:y2, x1:x2] += 1

        assert (count_mat == 0).sum() == 0
        seg_preds = seg_preds / count_mat
        if return_edge:
            edge_preds = edge_preds / count_mat

        if rescale:
            seg_preds = resize(
                seg_preds,
                size=img_meta[0]["ori_shape"][:2],
                mode="bilinear",
                align_corners=self.align_corners,
                warning=False,
            )
            if return_edge:
                edge_preds = resize(
                    edge_preds,
                    size=img_meta[0]["ori_shape"][:2],
                    mode="bilinear",
                    align_corners=self.align_corners,
                    warning=False,
                )

        if return_edge:
            return (seg_preds, edge_preds)
        else:
            return seg_preds



#!/usr/bin/env python3

"""Old losses that is converted for mmsegmentation

FIXME: update the losses so that we can reuse them
"""

import torch
import torch.nn as nn
import torch.nn.functional as F

from ..builder import SEG_LOSSES, JOINT_LOSSES


def calculate_weights(target, num_classes=19, norm=False, upper_bound=1.0):
    """
    Calculate weights of classes based on the training crop
    """

    bins = torch.histc(
        target,
        bins=num_classes,
        min=0.0,
        max=num_classes,
    )
    hist_norm = bins.float() / bins.sum()
    if norm:
        hist = ((bins != 0).float() * upper_bound * (1 / hist_norm)) + 1.0
    else:
        hist = ((bins != 0).float() * upper_bound * (1.0 - hist_norm)) + 1.0
    return hist


@SEG_LOSSES.register_module()
class ImageBasedCrossEntropyLoss2d(nn.Module):
    """
    Image Weighted Cross Entropy Loss
    """

    def __init__(
        self,
        num_classes,
        norm=False,
        upper_bound=1.0,
        loss_weight=1.0,
        loss_name="loss_imgwt_ce",
    ):
        super(ImageBasedCrossEntropyLoss2d, self).__init__()
        self.num_classes = num_classes
        self.norm = norm
        self.upper_bound = upper_bound
        self.loss_weight = loss_weight
        self._loss_name = loss_name

    def forward(
        self,
        cls_score,  # logits
        label,
        weight=None,
        ignore_index=255,
        **kwargs,
    ):
        # NOTE: calculate with batch
        weight = calculate_weights(
            label,
            num_classes=self.num_classes,
            norm=self.norm,
            upper_bound=self.upper_bound,
        )

        # I think this works since cls_score and labels are batches
        return self.loss_weight * F.cross_entropy(
            cls_score, label, weight=weight, reduction="mean", ignore_index=ignore_index
        )

    @property
    def loss_name(self):
        return self._loss_name


@JOINT_LOSSES.register_module()
class EdgeAttentionLoss(nn.Module):
    def __init__(
        self,
        num_classes,
        threshold=0.8,
        norm=False,
        upper_bound=1.0,
        loss_weight=1.0,
        loss_name="loss_edge_att",
    ):
        super().__init__()
        self.num_classes = num_classes
        assert 0 < threshold < 1
        self.threshold = threshold
        self.norm = norm
        self.upper_bound = upper_bound
        self.loss_weight = loss_weight
        self._loss_name = loss_name

    def forward(
        self,
        cls_score,  # logits
        label,
        edge,
        weight=None,
        ignore_index=255,
        **kwargs,
    ):
        filler = torch.ones_like(label) * 255
        weight = calculate_weights(
            label,
            num_classes=self.num_classes,
            norm=self.norm,
            upper_bound=self.upper_bound,
        )

        # turn edge to probability
        edge = torch.sigmoid(edge)

        assert len(edge.shape) == 4, f"edge shape should be 4dim, {edge.shape}"

        return self.loss_weight * F.cross_entropy(
            cls_score,
            torch.where(edge.max(1)[0] > self.threshold, label, filler),  # binary edge
            weight=weight,
            reduction="mean",
            ignore_index=ignore_index,
        )

    @property
    def loss_name(self):
        return self._loss_name



#!/usr/bin/env python3

from .seg import multi_gpu_seg_test, single_gpu_seg_test
from .edge import multi_gpu_edge_test, single_gpu_edge_test

__all__ = [
    "multi_gpu_seg_test",
    "single_gpu_seg_test",
    "multi_gpu_edge_test",
    "single_gpu_edge_test",
]



#!/usr/bin/env python3

from mmcv.utils import Registry
from mmseg.models.builder import (
    MODELS,
    NECKS,
    HEADS,
    LOSSES as SEG_LOSSES,
    SEGMENTORS,
    build_segmentor,
)
from blette.models.builder import (
    BACKBONES,
    LOSSES as EDGE_LOSSES,
)

POTATO_MODELS = Registry("potato_models", parent=MODELS)
JOINT_LOSSES = POTATO_MODELS

__all__ = [
    "BACKBONES",
    "NECKS",
    "HEADS",
    "SEG_LOSSES",
    "EDGE_LOSSES",
    "JOINT_LOSSES",
    "SEGMENTORS",
    "build_backbone",
    "build_neck",
    "build_head",
    "build_seg_loss",
    "build_edge_loss",
    "build_joint_loss",
    "build_segmentor",
]


def build_backbone(cfg):
    """Build backbone."""
    return BACKBONES.build(cfg)


def build_neck(cfg):
    """Build neck."""
    return NECKS.build(cfg)


def build_head(cfg):
    """Build head."""
    return HEADS.build(cfg)


def build_seg_loss(cfg):
    """Build segmentation loss."""
    return SEG_LOSSES.build(cfg)


def build_edge_loss(cfg):
    """Build edge loss."""
    return EDGE_LOSSES.build(cfg)


def build_joint_loss(cfg):
    """Build joint task loss"""
    return JOINT_LOSSES.build(cfg)



#!/usr/bin/env python3

"""Mask Gradient Losses

FIXME:
- maybe calculate masks in the loss instead (make it a joint loss)
"""

import torch
import torch.nn as nn
import torch.nn.functional as F

from ..builder import EDGE_LOSSES


@EDGE_LOSSES.register_module()
class DualityLoss(nn.Module):
    """Mask Gradient L1 Loss for RPCNet"""

    def __init__(
        self,
        loss_weight=1.0,
        loss_name="duality_l1",
        reduction="sum",
    ):
        super().__init__()
        self.loss_weight = loss_weight
        self._loss_name = loss_name
        self._reduction = reduction

    def forward(
        self,
        edge_logit,
        maskgrad,
        **kwargs,
    ):
        edge = torch.sigmoid(edge_logit)
        loss = F.l1_loss(
            input=maskgrad,
            target=edge,
            reduction=self._reduction,
        )
        if self._reduction == "sum":
            loss = loss / edge.numel()
        return self.loss_weight * loss

    @property
    def loss_name(self):
        return self._loss_name


@EDGE_LOSSES.register_module()
class MaskGradientCrossEntropyLoss(nn.Module):
    """FIXME: maskgrads are not logits

    I don't this this loss is working correctly
    """

    def __init__(
        self,
        loss_weight=1.0,
        loss_name="loss_mask_grad_ce",
    ):
        super().__init__()
        self.loss_weight = loss_weight
        self._loss_name = loss_name

    def forward(
        self,
        edge_logit,  # logits
        edge_label,
        weight=None,
        **kwargs,
    ):
        # FIXME: input segmentation instead of edge
        # FIXME: input
        loss_total = 0

        # FIXME: could optimize for batched loss
        for i in range(edge_label.size(0)):  # iterate for batch size
            pred = edge_logit[i]
            target = edge_label[i]

            num_pos = torch.sum(target)  # true positive number
            num_total = target.size(-1) * target.size(-2)  # true total number
            num_neg = num_total - num_pos
            # compute a pos_weight for each image
            pos_weight = (num_neg / num_pos).clamp(min=1, max=num_total)

            max_val = (-pred).clamp_min_(0)
            log_weight = 1 + (pos_weight - 1) * target
            loss = (
                pred
                - pred * target
                + log_weight
                * (
                    max_val
                    + torch.log(torch.exp(-max_val) + torch.exp(-pred - max_val))
                )
            )

            loss = loss.mean()
            loss_total = loss_total + loss

        loss_total = loss_total / edge_label.size(0)
        return self.loss_weight * loss_total

    @property
    def loss_name(self):
        return self._loss_name



#!/usr/bin/env python3

from .builder import DATASETS
from .custom import CustomJointDataset, OTFCustomJointDataset

ISPRS_CLASSES = (
    "impervious_surface",
    "building",
    "low_vegetation",
    "tree",
    "car",
    "clutter",
)
ISPRS_PALETTE = [
    [255, 255, 255],
    [0, 0, 255],
    [0, 255, 255],
    [0, 255, 0],
    [255, 255, 0],
    [255, 0, 0],
]
ISPRS_labelIds = [0, 1, 2, 3, 4, 5, 255]
# ISPRS_label2trainId = {
#     1: 0,
#     2: 1,
#     3: 2,
#     4: 3,
#     5: 4,
#     6: 5,
# }
ISPRS_label2trainId = {
    0: 0,
    1: 1,
    2: 2,
    3: 3,
    4: 4,
    5: 5,
}


@DATASETS.register_module()
class JointISPRSDataset(CustomJointDataset):
    """Joint ISPRS dataset.
    In segmentation map annotation for LoveDA, 0 is the ignore index.
    ``reduce_zero_label`` should be set to True. The ``img_suffix`` and
    ``seg_map_suffix`` are both fixed to '.png'.
    """

    CLASSES = ISPRS_CLASSES
    PALETTE = ISPRS_PALETTE

    def __init__(
        self,
        thin=False,
        test_mode=False,
        **kwargs,
    ):

        if test_mode:
            if thin:
                edge_map_suffix = "_thin.png"
            else:
                edge_map_suffix = "_raw.png"
        else:
            edge_map_suffix = ".png"

        super(JointISPRSDataset, self).__init__(
            img_suffix=".png",
            edge_map_suffix=edge_map_suffix,
            seg_map_suffix=".png",
            reduce_zero_label=True,
            inst_sensitive=False,
            **kwargs,
        )


@DATASETS.register_module()
class OTFJointISPRSDataset(OTFCustomJointDataset):
    CLASSES = ISPRS_CLASSES
    PALETTE = ISPRS_PALETTE

    def __init__(self, **kwargs):
        super(OTFJointISPRSDataset, self).__init__(
            img_suffix=".png",
            seg_map_suffix=".png",
            reduce_zero_label=True,
            inst_sensitive=False,
            labelIds=ISPRS_labelIds,
            ignore_indices=[6],  # ignore 255
            label2trainId=ISPRS_label2trainId,
            **kwargs,
        )



#!/usr/bin/env python3

"""Inria Aerial Image Labeling Dataset

https://project.inria.fr/aerialimagelabeling/
"""

from .builder import DATASETS
from .custom import CustomJointDataset, OTFCustomBinaryJointDataset

INRIA_CLASSES = ("background", "building")
INRIA_PALETTE = [[0, 0, 0], [255, 255, 255]]
INRIA_labelIds = [0, 1]


@DATASETS.register_module()
class INRIAAerialDataset(CustomJointDataset):

    CLASSES = INRIA_CLASSES
    PALETTE = INRIA_PALETTE

    def __init__(
        self,
        img_suffix=".png",
        seg_map_suffix=".png",
        edge_map_suffix=".png",
        gt_edge_loader_cfg=dict(binary=True),
        **kwargs,
    ):
        super().__init__(
            img_suffix=img_suffix,
            seg_map_suffix=seg_map_suffix,
            edge_map_suffix=edge_map_suffix,
            inst_sensitive=False,
            gt_edge_loader_cfg=gt_edge_loader_cfg,
            reduce_zero_label=False,  # zero is background!
            **kwargs,
        )


@DATASETS.register_module()
class OTFINRIAAerialDataset(OTFCustomBinaryJointDataset):

    CLASSES = INRIA_CLASSES
    PALETTE = INRIA_PALETTE
    IDS = [0, 1]

    def __init__(
        self,
        img_suffix=".png",
        seg_map_suffix=".png",
        reduce_zero_label=False,
        ignore_indices=[],
        labelIds=INRIA_labelIds,
        label2trainId=None,
        radius=1,
        selected_label=1,
        **kwargs,
    ):
        super().__init__(
            img_suffix=img_suffix,
            seg_map_suffix=seg_map_suffix,
            reduce_zero_label=reduce_zero_label,
            ignore_indices=ignore_indices,
            labelIds=labelIds,
            label2trainId=label2trainId,
            radius=radius,
            selected_label=selected_label,
            **kwargs,
        )



#!/usr/bin/env python3

import torch
import torch.nn as nn

from mmcv.cnn import ConvModule, DepthwiseSeparableConvModule
from mmseg.ops import resize
from mmseg.models.decode_heads.sep_aspp_head import DepthwiseSeparableASPPModule
from blette.models.utils import (
    GroupedConvFuse,
    LocationAdaptiveLearner,
    BasicBlockSideConv,
    SideConv,
)

from potato.models.builder import HEADS
from .manyedge_decode_head import BaseManyEdgeDecodeHead


@HEADS.register_module()
class DDSMHJSBHead(BaseManyEdgeDecodeHead):
    def __init__(
        self,
        dilations=(1, 6, 12, 18),
        c1_in_channels=256,
        c1_channels=48,
        num_side_blocks=2,
        edge_key="fuse",
        log_edge_keys=("fuse", "side1", "side2", "side3", "side4", "side5"),
        binary_edge_keys=("side1", "side2", "side3", "side4"),
        multilabel_edge_keys=("fuse", "side5"),
        loss_binary_edge=dict(type="BinaryEdgeLoss", loss_weight=1.0),
        loss_multilabel_edge=dict(type="MultiLabelEdgeLoss", loss_weight=1.0),
        feature_resize_index=-1,
        **kwargs,
    ):
        super().__init__(
            input_transform="multiple_select",
            edge_key=edge_key,
            log_edge_keys=log_edge_keys,
            binary_edge_keys=binary_edge_keys,
            multilabel_edge_keys=multilabel_edge_keys,
            loss_binary_edge=loss_binary_edge,
            loss_multilabel_edge=loss_multilabel_edge,
            **kwargs,
        )

        # changes which feature to `resize_to`
        self.feature_resize_index = feature_resize_index

        _interp = "bilinear"
        _bias = True

        # 1. Multi-label edge stream
        self.side1 = BasicBlockSideConv(
            in_channels=self.in_channels[0],
            out_channels=1,
            num_blocks=num_side_blocks,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            bias=_bias,
            act_cfg=self.act_cfg,
            interpolation=_interp,
            align_corners=self.align_corners,
        )
        self.side2 = BasicBlockSideConv(
            in_channels=self.in_channels[1],
            out_channels=1,
            num_blocks=num_side_blocks,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            bias=_bias,
            act_cfg=self.act_cfg,
            interpolation=_interp,
            align_corners=self.align_corners,
        )
        self.side3 = BasicBlockSideConv(
            in_channels=self.in_channels[2],
            out_channels=1,
            num_blocks=num_side_blocks,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            bias=_bias,
            act_cfg=self.act_cfg,
            interpolation=_interp,
            align_corners=self.align_corners,
        )
        self.side4 = BasicBlockSideConv(
            in_channels=self.in_channels[3],
            out_channels=1,
            num_blocks=num_side_blocks,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            bias=_bias,
            act_cfg=self.act_cfg,
            interpolation=_interp,
            align_corners=self.align_corners,
        )
        self.side5 = BasicBlockSideConv(
            in_channels=self.in_channels[4],
            out_channels=self.num_classes,
            num_blocks=num_side_blocks,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            bias=_bias,
            act_cfg=self.act_cfg,
            interpolation=_interp,
            align_corners=self.align_corners,
        )
        self.fuse = GroupedConvFuse(
            num_classes=self.num_classes,
            num_sides=5,
            conv_cfg=self.conv_cfg,
        )

        self.image_pool = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            ConvModule(
                self.in_channels[4],
                self.channels,
                1,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
            ),
        )
        self.aspp_modules = DepthwiseSeparableASPPModule(
            in_channels=self.in_channels[4],
            channels=self.channels,
            dilations=dilations,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )
        self.c1_bottleneck = ConvModule(
            in_channels=c1_in_channels,
            out_channels=c1_channels,
            kernel_size=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )

        self.bottleneck = ConvModule(
            (len(dilations) + 1) * self.channels,
            self.channels,
            kernel_size=3,
            padding=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )
        self.sep_bottleneck = nn.Sequential(
            DepthwiseSeparableConvModule(
                in_channels=self.channels + c1_channels,
                out_channels=self.channels,
                kernel_size=3,
                padding=1,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
            ),
            DepthwiseSeparableConvModule(
                in_channels=self.channels,
                out_channels=self.channels,
                kernel_size=3,
                padding=1,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
            ),
        )

    def forward(self, inputs):
        """Forward function"""

        # x = self._transform_inputs(inputs)  # out: list
        x = [i for i in inputs]
        # [stem, layer1, layer2, layer3, layer4, input_image]
        # h: 1/2, 1/4, 1/8, 1/8, 1/8, 1

        assert isinstance(x, list)
        bs, c, h, w = x[self.feature_resize_index].shape
        resize_to = (h, w)  # TODO: might be too large

        # Semantic Edge Stream
        side1 = self.side1(x[0], resize_to)  # (B, 1, H, W)
        side2 = self.side2(x[1], resize_to)  # (B, 1, H, W)
        side3 = self.side3(x[2], resize_to)  # (B, 1, H, W)
        side4 = self.side4(x[3], resize_to)  # (B, 1, H, W)
        side5 = self.side5(x[4], resize_to)  # (B, 19, H, W)
        fused_edges = self.fuse([side1, side2, side3, side4, side5])

        # refine aspp with edge features
        aspp_outs = [
            resize(
                self.image_pool(x[4]),
                size=x[4].size()[2:],
                mode="bilinear",
                align_corners=self.align_corners,
            )
        ]
        aspp_outs.extend(self.aspp_modules(x[4]))
        aspp_outs = torch.cat(aspp_outs, dim=1)
        output = self.bottleneck(aspp_outs)
        c1_output = self.c1_bottleneck(x[1])
        output = resize(
            input=output,
            size=c1_output.shape[2:],
            mode="bilinear",
            align_corners=self.align_corners,
        )

        output = torch.cat([output, c1_output], dim=1)
        output = self.sep_bottleneck(output)
        output = self.cls_seg(output)

        return (
            output,
            dict(side1=side1, side2=side2, side3=side3, side4=side4),
            # dict(side1=side1, side2=side2, side3=side3),
            dict(fuse=fused_edges, side5=side5),
        )


@HEADS.register_module()
class CASENetMHJSBHead(BaseManyEdgeDecodeHead):
    def __init__(
        self,
        dilations=(1, 6, 12, 18),
        c1_in_channels=256,
        c1_channels=48,
        edge_key="fuse",
        log_edge_keys=("fuse", "side5"),
        multilabel_edge_keys=("fuse", "side5"),
        loss_multilabel_edge=dict(type="MultiLabelEdgeLoss", loss_weight=1.0),
        feature_resize_index=-1,
        **kwargs,
    ):
        super().__init__(
            input_transform="multiple_select",
            edge_key=edge_key,
            log_edge_keys=log_edge_keys,
            multilabel_edge_keys=multilabel_edge_keys,
            loss_multilabel_edge=loss_multilabel_edge,
            **kwargs,
        )

        # changes which feature to `resize_to`
        self.feature_resize_index = feature_resize_index

        _interp = "bilinear"  # nearest
        _bias = False

        # bias should not be turn on when some of the sides are not supervised

        # Sides 1, 2, 3, 5
        self.side1 = SideConv(
            in_channels=self.in_channels[0],
            out_channels=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,  # bias instead of bn
            bias=_bias,  # add bias in the last layer
            act_cfg=self.act_cfg,
            interpolation=_interp,
            align_corners=self.align_corners,
        )
        self.side2 = SideConv(
            in_channels=self.in_channels[1],
            out_channels=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,  # bias instead of bn
            bias=_bias,
            act_cfg=self.act_cfg,
            interpolation=_interp,
            align_corners=self.align_corners,
        )
        self.side3 = SideConv(
            in_channels=self.in_channels[2],
            out_channels=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,  # bias instead of bn
            bias=_bias,
            act_cfg=self.act_cfg,
            interpolation=_interp,
            align_corners=self.align_corners,
        )
        self.side5 = SideConv(
            in_channels=self.in_channels[4],
            out_channels=self.num_classes,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,  # bias instead of bn
            bias=_bias,
            act_cfg=self.act_cfg,
            interpolation=_interp,
            align_corners=self.align_corners,
        )
        self.fuse = GroupedConvFuse(
            num_classes=self.num_classes,
            num_sides=4,
            conv_cfg=self.conv_cfg,
            bias=_bias,  # originally true
        )

        self.image_pool = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            ConvModule(
                self.in_channels[4],
                self.channels,
                1,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
            ),
        )
        self.aspp_modules = DepthwiseSeparableASPPModule(
            in_channels=self.in_channels[4],
            channels=self.channels,
            dilations=dilations,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )
        self.c1_bottleneck = ConvModule(
            in_channels=c1_in_channels,
            out_channels=c1_channels,
            kernel_size=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )

        self.bottleneck = ConvModule(
            (len(dilations) + 1) * self.channels,
            self.channels,
            kernel_size=3,
            padding=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )
        self.sep_bottleneck = nn.Sequential(
            DepthwiseSeparableConvModule(
                in_channels=self.channels + c1_channels,
                out_channels=self.channels,
                kernel_size=3,
                padding=1,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
            ),
            DepthwiseSeparableConvModule(
                in_channels=self.channels,
                out_channels=self.channels,
                kernel_size=3,
                padding=1,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
            ),
        )

    def forward(self, inputs):
        """Forward function"""

        # x = self._transform_inputs(inputs)  # out: list
        x = [i for i in inputs]
        # [stem, layer1, layer2, layer3, layer4, input_image]
        # h: 1/2, 1/4, 1/8, 1/8, 1/8, 1

        assert isinstance(x, list)
        bs, c, h, w = x[self.feature_resize_index].shape
        resize_to = (h, w)  # TODO: might be too large

        # Semantic Edge Stream
        side1 = self.side1(x[0], resize_to)  # (B, 1, H, W)
        side2 = self.side2(x[1], resize_to)  # (B, 1, H, W)
        side3 = self.side3(x[2], resize_to)  # (B, 1, H, W)
        side5 = self.side5(x[4], resize_to)  # (B, 19, H, W)
        fused_edges = self.fuse([side1, side2, side3, side5])

        # refine aspp with edge features
        aspp_outs = [
            resize(
                self.image_pool(x[4]),
                size=x[4].size()[2:],
                mode="bilinear",
                align_corners=self.align_corners,
            )
        ]
        aspp_outs.extend(self.aspp_modules(x[4]))
        aspp_outs = torch.cat(aspp_outs, dim=1)
        output = self.bottleneck(aspp_outs)
        c1_output = self.c1_bottleneck(x[1])
        output = resize(
            input=output,
            size=c1_output.shape[2:],
            mode="bilinear",
            align_corners=self.align_corners,
        )

        output = torch.cat([output, c1_output], dim=1)
        output = self.sep_bottleneck(output)
        output = self.cls_seg(output)

        return (
            output,
            dict(side1=side1, side2=side2, side3=side3),
            dict(fuse=fused_edges, side5=side5),
        )


@HEADS.register_module()
class DFFMHJSBHead(BaseManyEdgeDecodeHead):
    def __init__(
        self,
        dilations=(1, 6, 12, 18),
        c1_in_channels=256,
        c1_channels=48,
        edge_key="fuse",
        log_edge_keys=("fuse", "side5"),
        multilabel_edge_keys=("fuse", "side5"),
        loss_multilabel_edge=dict(type="MultiLabelEdgeLoss", loss_weight=1.0),
        feature_resize_index=-1,
        **kwargs,
    ):
        super().__init__(
            input_transform="multiple_select",
            edge_key=edge_key,
            log_edge_keys=log_edge_keys,
            multilabel_edge_keys=multilabel_edge_keys,
            loss_multilabel_edge=loss_multilabel_edge,
            **kwargs,
        )

        # changes which feature to `resize_to`
        self.feature_resize_index = feature_resize_index

        _interp = "bilinear"  # nearest
        _bias = True  # turn on?

        # Sides 1, 2, 3, 5 and Side 5 Weight
        self.side1 = SideConv(
            in_channels=self.in_channels[0],
            out_channels=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            bias=_bias,
            act_cfg=self.act_cfg,
            interpolation=_interp,
            align_corners=self.align_corners,
        )
        self.side2 = SideConv(
            in_channels=self.in_channels[1],
            out_channels=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            bias=_bias,
            act_cfg=self.act_cfg,
            interpolation=_interp,
            align_corners=self.align_corners,
        )
        self.side3 = SideConv(
            in_channels=self.in_channels[2],
            out_channels=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            bias=_bias,
            act_cfg=self.act_cfg,
            interpolation=_interp,
            align_corners=self.align_corners,
        )
        self.side5 = SideConv(
            in_channels=self.in_channels[4],
            out_channels=self.num_classes,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            bias=_bias,
            act_cfg=self.act_cfg,
            interpolation=_interp,
            align_corners=self.align_corners,
        )
        self.side5_w = SideConv(
            in_channels=self.in_channels[4],
            out_channels=self.num_classes * 4,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            bias=_bias,  # might not need?
            act_cfg=self.act_cfg,
            interpolation=_interp,
            align_corners=self.align_corners,
        )
        self.ada_learner = LocationAdaptiveLearner(
            in_channels=self.num_classes * 4,
            out_channels=self.num_classes * 4,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )

        self.image_pool = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            ConvModule(
                self.in_channels[4],
                self.channels,
                1,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
            ),
        )
        self.aspp_modules = DepthwiseSeparableASPPModule(
            in_channels=self.in_channels[4],
            channels=self.channels,
            dilations=dilations,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )
        self.c1_bottleneck = ConvModule(
            in_channels=c1_in_channels,
            out_channels=c1_channels,
            kernel_size=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )

        self.bottleneck = ConvModule(
            (len(dilations) + 1) * self.channels,
            self.channels,
            kernel_size=3,
            padding=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )
        self.sep_bottleneck = nn.Sequential(
            DepthwiseSeparableConvModule(
                in_channels=self.channels + c1_channels,
                out_channels=self.channels,
                kernel_size=3,
                padding=1,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
            ),
            DepthwiseSeparableConvModule(
                in_channels=self.channels,
                out_channels=self.channels,
                kernel_size=3,
                padding=1,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
            ),
        )

    def forward(self, inputs):
        """Forward function"""

        # x = self._transform_inputs(inputs)  # out: list
        x = [i for i in inputs]
        # [stem, layer1, layer2, layer3, layer4, input_image]
        # h: 1/2, 1/4, 1/8, 1/8, 1/8, 1

        assert isinstance(x, list)
        bs, c, h, w = x[self.feature_resize_index].shape
        resize_to = (h, w)  # TODO: might be too large

        # Semantic Edge Side Stream
        side1 = self.side1(x[0], resize_to)  # (B, 1, H, W)
        side2 = self.side2(x[1], resize_to)  # (B, 1, H, W)
        side3 = self.side3(x[2], resize_to)  # (B, 1, H, W)
        side5 = self.side5(x[4], resize_to)  # (B, 19, H, W)
        side5_w = self.side5_w(x[4], resize_to)  # (B, 19*4, H, W)
        fused_edges = self.ada_learner([side1, side2, side3, side5, side5_w])

        # refine aspp with edge features
        aspp_outs = [
            resize(
                self.image_pool(x[4]),
                size=x[4].size()[2:],
                mode="bilinear",
                align_corners=self.align_corners,
            )
        ]
        aspp_outs.extend(self.aspp_modules(x[4]))
        aspp_outs = torch.cat(aspp_outs, dim=1)
        output = self.bottleneck(aspp_outs)
        c1_output = self.c1_bottleneck(x[1])
        output = resize(
            input=output,
            size=c1_output.shape[2:],
            mode="bilinear",
            align_corners=self.align_corners,
        )

        output = torch.cat([output, c1_output], dim=1)
        output = self.sep_bottleneck(output)
        output = self.cls_seg(output)

        return (
            output,
            dict(side1=side1, side2=side2, side3=side3),
            dict(fuse=fused_edges, side5=side5),
        )



#!/usr/bin/env python3

"""Boundary F-score evaluation

NOTE: only single gpu evaluation is supported

TODO: support multi-gpu or distributed
"""

import torch

import mmcv


def single_gpu_boundary_fscore_test(
    model,
    data_loader,
    bound_th=0.008,
    num_proc=16,
):
    model.eval()
    results = []
    dataset = data_loader.dataset
    prog_bar = mmcv.ProgressBar(len(dataset))

    # The pipeline about how the data_loader retrieval samples from dataset:
    # sampler -> batch_sampler -> indices
    # The indices are passed to dataset_fetcher to get data from dataset.
    # data_fetcher -> collate_fn(dataset[index]) -> data_sample
    # we use batch_sampler to get correct data idx
    loader_indices = data_loader.batch_sampler

    for batch_indices, data in zip(loader_indices, data_loader):

        with torch.no_grad():
            result = model(return_loss=False, **data)

        result = dataset.pre_boundary_fscore_eval(
            result,
            indices=batch_indices,
            bound_th=bound_th,
            num_proc=num_proc,
        )
        results.extend(result)

        batch_size = len(result)
        for _ in range(batch_size):
            prog_bar.update()

    return results



from .sep_aspp_grad_head import GradASPPHead
from .gscnn_head import GSCNNHead, MultiLabelGSCNNHead
from .dense_aspp_head import DenseASPPHead
from .rpc_head import RPCHead

from .mhjsb_fcn import (
    DDSFCNMHJSBHead,
    DFFFCNMHJSBHead,
    CASENetFCNMHJSBHead,
)
from .mhjsb_pspnet import (
    DDSPSPNetMHJSBHead,
    DFFPSPNetMHJSBHead,
    CASENetPSPNetMHJSBHead,
)
from .mhjsb_deeplabv3plus import (
    DDSMHJSBHead,
    DFFMHJSBHead,
    CASENetMHJSBHead,
)
from .ts_jsb import TSJSBHead
from .ts_jsbv1 import (
    DDSTSJSBHead,
    DFFTSJSBHead,
    CASENetTSJSBHead,
)
from .ts_jsbv2 import CASENetTSJSBV2Head

# from .debug_gscnn_head import MultiLabelGSCNNHead
from .three_stream_head import ThreeStreamASPPHead

# experimental
from .binary_hed_head import BinaryHEDHead

__all__ = [
    "GradASPPHead",
    "GSCNNHead",
    "MultiLabelGSCNNHead",
    "DenseASPPHead",
    "RPCHead",
    "DDSFCNMHJSBHead",
    "DFFFCNMHJSBHead",
    "CASENetFCNMHJSBHead",
    "DDSPSPNetMHJSBHead",
    "DFFPSPNetMHJSBHead",
    "CASENetPSPNetMHJSBHead",
    "DDSMHJSBHead",
    "DFFMHJSBHead",
    "CASENetMHJSBHead",
    "TSJSBHead",
    "DDSTSJSBHead",
    "DFFTSJSBHead",
    "CASENetTSJSBHead",
    "CASENetTSJSBV2Head",
    "ThreeStreamASPPHead",
    "BinaryHEDHead",
]



#!/usr/bin/env python3

"""
NOTE
- since `reduce_zero_label=True`, we have to be careful about what we pass to Mask2Edge
"""

import os.path as osp

import mmcv
import numpy as np
from PIL import Image

from .builder import DATASETS
from .custom import CustomJointDataset, OTFCustomJointDataset

LOVEDA_CLASSES = (
    "background",
    "building",
    "road",
    "water",
    "barren",
    "forest",
    "agricultural",
)
LOVEDA_PALETTE = [
    [255, 255, 255],
    [255, 0, 0],
    [255, 255, 0],
    [0, 0, 255],
    [159, 129, 183],
    [0, 255, 0],
    [255, 195, 128],
]
LOVEDA_labelIds = [0, 1, 2, 3, 4, 5, 6, 255]
LOVEDA_label2trainId = {
    0: 0,
    1: 1,
    2: 2,
    3: 3,
    4: 4,
    5: 5,
    6: 6,
}


@DATASETS.register_module()
class JointLoveDADataset(CustomJointDataset):
    """LoveDA dataset.
    In segmentation map annotation for LoveDA, 0 is the ignore index.
    ``reduce_zero_label`` should be set to True. The ``img_suffix`` and
    ``seg_map_suffix`` are both fixed to '.png'.
    """

    CLASSES = LOVEDA_CLASSES
    PALETTE = LOVEDA_PALETTE

    def __init__(
        self,
        thin=False,
        test_mode=False,
        **kwargs,
    ):
        if test_mode:
            if thin:
                edge_map_suffix = "_thin.png"
            else:
                edge_map_suffix = "_raw.png"
        else:
            edge_map_suffix = ".png"

        super(JointLoveDADataset, self).__init__(
            img_suffix=".png",
            edge_map_suffix=edge_map_suffix,
            seg_map_suffix=".png",
            reduce_zero_label=True,
            inst_sensitive=False,
            **kwargs,
        )

    def results2img(self, results, imgfile_prefix, indices=None):
        """Write the segmentation results to images.
        Args:
            results (list[ndarray]): Testing results of the
                dataset.
            imgfile_prefix (str): The filename prefix of the png files.
                If the prefix is "somepath/xxx",
                the png files will be named "somepath/xxx.png".
            indices (list[int], optional): Indices of input results, if not
                set, all the indices of the dataset will be used.
                Default: None.
        Returns:
            list[str: str]: result txt files which contains corresponding
            semantic segmentation images.
        """

        mmcv.mkdir_or_exist(imgfile_prefix)
        result_files = []
        for result, idx in zip(results, indices):

            filename = self.img_infos[idx]["filename"]
            basename = osp.splitext(osp.basename(filename))[0]

            png_filename = osp.join(imgfile_prefix, f"{basename}.png")

            # The  index range of official requirement is from 0 to 6.
            output = Image.fromarray(result.astype(np.uint8))
            output.save(png_filename)
            result_files.append(png_filename)

        return result_files

    def format_results(self, results, imgfile_prefix, indices=None):
        """Format the results into dir (standard format for LoveDA evaluation).
        Args:
            results (list): Testing results of the dataset.
            imgfile_prefix (str): The prefix of images files. It
                includes the file path and the prefix of filename, e.g.,
                "a/b/prefix".
            indices (list[int], optional): Indices of input results,
                if not set, all the indices of the dataset will be used.
                Default: None.
        Returns:
            tuple: (result_files, tmp_dir), result_files is a list containing
                the image paths, tmp_dir is the temporal directory created
                for saving json/png files when img_prefix is not specified.
        """
        if indices is None:
            indices = list(range(len(self)))

        assert isinstance(results, list), "results must be a list."
        assert isinstance(indices, list), "indices must be a list."

        result_files = self.results2img(results, imgfile_prefix, indices)

        return result_files


@DATASETS.register_module()
class OTFJointLoveDADataset(OTFCustomJointDataset):
    CLASSES = LOVEDA_CLASSES
    PALETTE = LOVEDA_PALETTE

    def __init__(
        self,
        **kwargs,
    ):
        super(OTFJointLoveDADataset, self).__init__(
            img_suffix=".png",
            seg_map_suffix=".png",
            reduce_zero_label=True,
            inst_sensitive=False,
            labelIds=LOVEDA_labelIds,
            ignore_indices=[7],  # ignore 255
            label2trainId=LOVEDA_label2trainId,
            **kwargs,
        )

    def format_results(self, **kwargs):
        raise ValueError("ERR: should not use OTF for testing")



#!/usr/bin/env python3

import os.path as osp
import tempfile
import warnings

import numpy as np
import torch

import mmcv
from mmcv.engine import collect_results_cpu, collect_results_gpu
from mmcv.image import tensor2imgs
from mmcv.runner import get_dist_info


def np2tmp(array, temp_file_name=None, tmpdir=None):
    """Save ndarray to local numpy file.

    Args:
        array (ndarray): Ndarray to save.
        temp_file_name (str): Numpy file name. If 'temp_file_name=None', this
            function will generate a file name with tempfile.NamedTemporaryFile
            to save ndarray. Default: None.
        tmpdir (str): Temporary directory to save Ndarray files. Default: None.
    Returns:
        str: The numpy file name.
    """

    if temp_file_name is None:
        temp_file_name = tempfile.NamedTemporaryFile(
            suffix=".npy", delete=False, dir=tmpdir
        ).name
    np.save(temp_file_name, array)
    return temp_file_name


def single_gpu_seg_test(
    model,
    data_loader,
    show=False,
    out_dir=None,
    efficient_test=False,
    opacity=0.5,
    pre_eval=False,
    format_only=False,
    format_args={},
):
    """Test with single GPU by progressive mode.

    Args:
        model (nn.Module): Model to be tested.
        data_loader (utils.data.Dataloader): Pytorch data loader.
        show (bool): Whether show results during inference. Default: False.
        out_dir (str, optional): If specified, the results will be dumped into
            the directory to save output results.
        efficient_test (bool): Whether save the results as local numpy files to
            save CPU memory during evaluation. Mutually exclusive with
            pre_eval and format_results. Default: False.
        opacity(float): Opacity of painted segmentation map.
            Default 0.5.
            Must be in (0, 1] range.
        pre_eval (bool): Use dataset.pre_eval() function to generate
            pre_results for metric evaluation. Mutually exclusive with
            efficient_test and format_results. Default: False.
        format_only (bool): Only format result for results commit.
            Mutually exclusive with pre_eval and efficient_test.
            Default: False.
        format_args (dict): The args for format_results. Default: {}.
    Returns:
        list: list of evaluation pre-results or list of save file names.
    """
    if efficient_test:
        warnings.warn(
            "DeprecationWarning: ``efficient_test`` will be deprecated, the "
            "evaluation is CPU memory friendly with pre_eval=True"
        )
        mmcv.mkdir_or_exist(".efficient_test")
    # when none of them is set true, return segmentation results as
    # a list of np.array.
    assert [efficient_test, pre_eval, format_only].count(True) <= 1, (
        "``efficient_test``, ``pre_eval`` and ``format_only`` are mutually "
        "exclusive, only one of them could be true ."
    )

    model.eval()
    seg_results = []
    dataset = data_loader.dataset
    prog_bar = mmcv.ProgressBar(len(dataset))
    # The pipeline about how the data_loader retrieval samples from dataset:
    # sampler -> batch_sampler -> indices
    # The indices are passed to dataset_fetcher to get data from dataset.
    # data_fetcher -> collate_fn(dataset[index]) -> data_sample
    # we use batch_sampler to get correct data idx
    loader_indices = data_loader.batch_sampler

    for batch_indices, data in zip(loader_indices, data_loader):
        with torch.no_grad():
            segs = model(
                return_loss=False,
                return_edge=False,
                **data,
            )
        assert isinstance(segs, list), f"segs should be list but got {type(segs)}"

        if show or out_dir:
            img_tensor = data["img"][0]
            img_metas = data["img_metas"][0].data[0]
            imgs = tensor2imgs(img_tensor, **img_metas[0]["img_norm_cfg"])
            assert len(imgs) == len(img_metas)

            for i, (img, img_meta) in enumerate(zip(imgs, img_metas)):
                h, w, _ = img_meta["img_shape"]
                img_show = img[:h, :w, :]

                ori_h, ori_w = img_meta["ori_shape"][:-1]
                img_show = mmcv.imresize(img_show, (ori_w, ori_h))

                if out_dir:
                    out_file = osp.join(out_dir, img_meta["ori_filename"])
                else:
                    out_file = None

                model.module.show_seg_result(
                    img_show,
                    segs[i],
                    palette=dataset.PALETTE,
                    show=show,
                    out_file=out_file,
                    opacity=opacity,
                )

        if efficient_test:
            segs = [np2tmp(_, tmpdir=".efficient_test") for _ in segs]

        if format_only:
            # format segmentation results for evaluation in servers
            segs = dataset.format_results(segs, indices=batch_indices, **format_args)
        if pre_eval:
            # TODO: adapt samples_per_gpu > 1.
            # only samples_per_gpu=1 valid now
            segs = dataset.pre_eval(segs, indices=batch_indices)
            seg_results.extend(segs)
        else:
            seg_results.extend(segs)

        batch_size = len(segs)
        for _ in range(batch_size):
            prog_bar.update()

    return seg_results


def multi_gpu_seg_test(
    model,
    data_loader,
    tmpdir=None,
    gpu_collect=False,
    efficient_test=False,
    pre_eval=False,
    format_only=False,
    format_args={},
):
    """Test model with multiple gpus by progressive mode.

    This method tests model with multiple gpus and collects the results
    under two different modes: gpu and cpu modes. By setting 'gpu_collect=True'
    it encodes results to gpu tensors and use gpu communication for results
    collection. On cpu mode it saves the results on different gpus to 'tmpdir'
    and collects them by the rank 0 worker.

    Args:
        model (nn.Module): Model to be tested.
        data_loader (utils.data.Dataloader): Pytorch data loader.
        tmpdir (str): Path of directory to save the temporary results from
            different gpus under cpu mode. The same path is used for efficient
            test. Default: None.
        gpu_collect (bool): Option to use either gpu or cpu to collect results.
            Default: False.
        efficient_test (bool): Whether save the results as local numpy files to
            save CPU memory during evaluation. Mutually exclusive with
            pre_eval and format_results. Default: False.
        pre_eval (bool): Use dataset.pre_eval() function to generate
            pre_results for metric evaluation. Mutually exclusive with
            efficient_test and format_results. Default: False.
        format_only (bool): Only format result for results commit.
            Mutually exclusive with pre_eval and efficient_test.
            Default: False.
        format_args (dict): The args for format_results. Default: {}.

    Returns:
        list: list of evaluation pre-results or list of save file names.
    """
    if efficient_test:
        warnings.warn(
            "DeprecationWarning: ``efficient_test`` will be deprecated, the "
            "evaluation is CPU memory friendly with pre_eval=True"
        )
        mmcv.mkdir_or_exist(".efficient_test")
    # when none of them is set true, return segmentation results as
    # a list of np.array.
    assert [efficient_test, pre_eval, format_only].count(True) <= 1, (
        "``efficient_test``, ``pre_eval`` and ``format_only`` are mutually "
        "exclusive, only one of them could be true ."
    )

    model.eval()
    results = []
    dataset = data_loader.dataset
    # The pipeline about how the data_loader retrieval samples from dataset:
    # sampler -> batch_sampler -> indices
    # The indices are passed to dataset_fetcher to get data from dataset.
    # data_fetcher -> collate_fn(dataset[index]) -> data_sample
    # we use batch_sampler to get correct data idx

    # batch_sampler based on DistributedSampler, the indices only point to data
    # samples of related machine.
    loader_indices = data_loader.batch_sampler

    rank, world_size = get_dist_info()
    if rank == 0:
        prog_bar = mmcv.ProgressBar(len(dataset))

    for batch_indices, data in zip(loader_indices, data_loader):
        with torch.no_grad():
            result = model(return_loss=False, rescale=True, **data)

        if efficient_test:
            result = [np2tmp(_, tmpdir=".efficient_test") for _ in result]

        if format_only:
            result = dataset.format_results(
                result, indices=batch_indices, **format_args
            )
        if pre_eval:
            # TODO: adapt samples_per_gpu > 1.
            # only samples_per_gpu=1 valid now
            result = dataset.pre_eval(result, indices=batch_indices)

        results.extend(result)

        if rank == 0:
            batch_size = len(result) * world_size
            for _ in range(batch_size):
                prog_bar.update()

    # collect results from all ranks
    if gpu_collect:
        results = collect_results_gpu(results, len(dataset))
    else:
        results = collect_results_cpu(results, len(dataset), tmpdir)
    return results



#!/usr/bin/env python3

from .formatting import (
    FormatJoint,
    FormatJointBinaryEdge,
    JointFormatBundle,
    JointBinaryFormatBundle,
)

__all__ = [
    "FormatJoint",
    "FormatJointBinaryEdge",
    "JointFormatBundle",
    "JointBinaryFormatBundle",
]



#!/usr/bin/env python3

"""Wider ResNet Backbone

Reference:
https://github.com/haruishi43/GSCNN/blob/master/sseg/network/wider_resnet.py
"""

from collections import OrderedDict
from functools import partial

import torch.nn as nn
import torch

from mmcv.cnn import build_activation_layer, build_norm_layer
from mmcv.runner import BaseModule

from ..builder import BACKBONES


def bnrelu(
    channels,
    norm_cfg=dict(type="BN", requires_grad=True),
    act_cfg=dict(type="ReLU", inplace=True),
):
    return nn.Sequential(
        build_norm_layer(norm_cfg, channels)[1],
        build_activation_layer(act_cfg),
    )


class GlobalAvgPool2d(nn.Module):
    def __init__(self):
        """Global average pooling over the input's spatial dimensions"""
        super().__init__()

    def forward(self, inputs):
        in_size = inputs.size()
        return inputs.view((in_size[0], in_size[1], -1)).mean(dim=2)


class IdentityResidualBlock(nn.Module):
    def __init__(
        self,
        in_channels,
        channels,
        stride=1,
        dilation=1,
        groups=1,
        norm_cfg=dict(type="BN", requires_grad=True),
        act_cfg=dict(type="ReLU", inplace=True),
        dropout=None,
        dist_bn=False,
    ):
        """Configurable identity-mapping residual block
        Parameters
        ----------
        in_channels : int
            Number of input channels.
        channels : list of int
            Number of channels in the internal feature maps.
            Can either have two or three elements: if three construct
            a residual block with two `3 x 3` convolutions,
            otherwise construct a bottleneck block with `1 x 1`, then
            `3 x 3` then `1 x 1` convolutions.
        stride : int
            Stride of the first `3 x 3` convolution
        dilation : int
            Dilation to apply to the `3 x 3` convolutions.
        groups : int
            Number of convolution groups.
            This is used to create ResNeXt-style blocks and is only compatible with
            bottleneck blocks.
        norm_cfg : dict
        act_cfg : dict
        dropout: callable
            Function to create Dropout Module.
        dist_bn: Boolean
            A variable to enable or disable use of distributed BN
        """
        super().__init__()
        self.dist_bn = dist_bn

        # Check parameters for inconsistencies
        if len(channels) != 2 and len(channels) != 3:
            raise ValueError("channels must contain either two or three values")
        if len(channels) == 2 and groups != 1:
            raise ValueError("groups > 1 are only valid if len(channels) == 3")

        is_bottleneck = len(channels) == 3
        need_proj_conv = stride != 1 or in_channels != channels[-1]

        self.bn1 = bnrelu(in_channels, norm_cfg=norm_cfg, act_cfg=act_cfg)
        if not is_bottleneck:
            layers = [
                (
                    "conv1",
                    nn.Conv2d(
                        in_channels,
                        channels[0],
                        3,
                        stride=stride,
                        padding=dilation,
                        bias=False,
                        dilation=dilation,
                    ),
                ),
                ("bn2", bnrelu(channels[0], norm_cfg=norm_cfg, act_cfg=act_cfg)),
                (
                    "conv2",
                    nn.Conv2d(
                        channels[0],
                        channels[1],
                        3,
                        stride=1,
                        padding=dilation,
                        bias=False,
                        dilation=dilation,
                    ),
                ),
            ]
            if dropout is not None:
                layers = layers[0:2] + [("dropout", dropout())] + layers[2:]
        else:
            layers = [
                (
                    "conv1",
                    nn.Conv2d(
                        in_channels,
                        channels[0],
                        1,
                        stride=stride,
                        padding=0,
                        bias=False,
                    ),
                ),
                ("bn2", bnrelu(channels[0], norm_cfg=norm_cfg, act_cfg=act_cfg)),
                (
                    "conv2",
                    nn.Conv2d(
                        channels[0],
                        channels[1],
                        3,
                        stride=1,
                        padding=dilation,
                        bias=False,
                        groups=groups,
                        dilation=dilation,
                    ),
                ),
                ("bn3", bnrelu(channels[1], norm_cfg=norm_cfg, act_cfg=act_cfg)),
                (
                    "conv3",
                    nn.Conv2d(
                        channels[1], channels[2], 1, stride=1, padding=0, bias=False
                    ),
                ),
            ]
            if dropout is not None:
                layers = layers[0:4] + [("dropout", dropout())] + layers[4:]
        self.convs = nn.Sequential(OrderedDict(layers))

        if need_proj_conv:
            self.proj_conv = nn.Conv2d(
                in_channels, channels[-1], 1, stride=stride, padding=0, bias=False
            )

    def forward(self, x):
        """
        This is the standard forward function for non-distributed batch norm
        """
        if hasattr(self, "proj_conv"):
            bn1 = self.bn1(x)
            shortcut = self.proj_conv(bn1)
        else:
            shortcut = x.clone()
            bn1 = self.bn1(x)

        out = self.convs(bn1)
        out = out + shortcut
        return out


@BACKBONES.register_module()
class WiderResNet(BaseModule):

    arch_settings = {
        16: (1, 1, 1, 1, 1, 1),
        20: (1, 1, 1, 3, 1, 1),
        38: (3, 3, 6, 3, 1, 1),
    }

    def __init__(
        self,
        depth,
        out_indices=(0, 2, 3, 6),
        norm_cfg=dict(type="BN", requires_grad=True),
        act_cfg=dict(type="ReLU", inplace=True),
        init_cfg=None,
    ):
        """Wider ResNet with pre-activation (identity mapping) blocks Parameters

        ----------
        structure : list of int
            Number of residual blocks in each of the six modules of the network.
        norm_act : callable
            Function to create normalization / activation Module.
        classes : int
            If not `0` also include global average pooling and \
            a fully-connected layer with `classes` outputs at the end
            of the network.
        """
        super().__init__(init_cfg)
        assert max(out_indices) < 7
        self.out_indices = out_indices

        structure = self.arch_settings[depth]

        # Initial layers
        self.mod1 = nn.Sequential(
            OrderedDict(
                [("conv1", nn.Conv2d(3, 64, 3, stride=1, padding=1, bias=False))]
            )
        )

        # Groups of residual blocks
        in_channels = 64
        channels = [
            (128, 128),
            (256, 256),
            (512, 512),
            (512, 1024),
            (512, 1024, 2048),
            (1024, 2048, 4096),
        ]
        for mod_id, num in enumerate(structure):
            # Create blocks for module
            blocks = []
            for block_id in range(num):
                blocks.append(
                    (
                        "block%d" % (block_id + 1),
                        IdentityResidualBlock(
                            in_channels,
                            channels[mod_id],
                            norm_cfg=norm_cfg,
                            act_cfg=act_cfg,
                        ),
                    )
                )

                # Update channels and p_keep
                in_channels = channels[mod_id][-1]

            # Create module
            if mod_id <= 4:
                self.add_module(
                    "pool%d" % (mod_id + 2), nn.MaxPool2d(3, stride=2, padding=1)
                )
            self.add_module("mod%d" % (mod_id + 2), nn.Sequential(OrderedDict(blocks)))

        # Pooling and predictor
        # self.bn_out = bnrelu(in_channels, norm_cfg=norm_cfg, act_cfg=act_cfg)

    def forward(self, img):
        r1 = self.mod1(img)
        r2 = self.mod2(self.pool2(r1))
        r3 = self.mod3(self.pool3(r2))
        r4 = self.mod4(self.pool4(r3))
        r5 = self.mod5(self.pool5(r4))
        r6 = self.mod6(self.pool6(r5))
        r7 = self.mod7(r6)
        # out = self.bn_out(out)

        outs = [r1, r2, r3, r4, r5, r6, r7]
        outs = [outs[i] for i in self.out_indices]

        return outs


@BACKBONES.register_module()
class WiderResNetA2(BaseModule):

    arch_settings = {
        16: (1, 1, 1, 1, 1, 1),
        20: (1, 1, 1, 3, 1, 1),
        38: (3, 3, 6, 3, 1, 1),
    }

    def __init__(
        self,
        depth,
        out_indices=(0, 2, 3, 6),
        dilation=False,  # GSCNN: True
        dist_bn=False,
        norm_cfg=dict(type="BN", requires_grad=True),
        act_cfg=dict(type="ReLU", inplace=True),
        init_cfg=None,
    ):
        """Wider ResNet with pre-activation (identity mapping) blocks
        This variant uses down-sampling by max-pooling in the first two blocks and
        by strided convolution in the others.

        Parameters
        ----------
        structure : list of int
            Number of residual blocks in each of the six modules of the network.
        norm_cfg : dict
        act_cfg : dict
        classes : int
            If not `0` also include global average pooling and a fully-connected layer
            with `classes` outputs at the end
            of the network.
        dilation : bool
            If `True` apply dilation to the last three modules and change the
            down-sampling factor from 32 to 8.
        """
        super().__init__(init_cfg)

        assert max(out_indices) < 7
        self.out_indices = out_indices
        self.dist_bn = dist_bn

        structure = self.arch_settings[depth]
        self.dilation = dilation

        # Initial layers
        self.mod1 = torch.nn.Sequential(
            OrderedDict(
                [("conv1", nn.Conv2d(3, 64, 3, stride=1, padding=1, bias=False))]
            )
        )

        # Groups of residual blocks
        in_channels = 64
        channels = [
            (128, 128),
            (256, 256),
            (512, 512),
            (512, 1024),
            (512, 1024, 2048),
            (1024, 2048, 4096),
        ]
        for mod_id, num in enumerate(structure):
            # Create blocks for module
            blocks = []
            for block_id in range(num):

                # NOTE: for `a2` we set the number of dilation and strides accordingly

                if not dilation:
                    dil = 1
                    stride = 2 if block_id == 0 and 2 <= mod_id <= 4 else 1
                else:
                    if mod_id == 3:
                        dil = 2
                    elif mod_id > 3:
                        dil = 4
                    else:
                        dil = 1
                    stride = 2 if block_id == 0 and mod_id == 2 else 1

                # NOTE: there are also dropouts
                if mod_id == 4:
                    drop = partial(nn.Dropout2d, p=0.3)
                elif mod_id == 5:
                    drop = partial(nn.Dropout2d, p=0.5)
                else:
                    drop = None

                blocks.append(
                    (
                        "block%d" % (block_id + 1),
                        IdentityResidualBlock(
                            in_channels,
                            channels[mod_id],
                            norm_cfg=norm_cfg,
                            act_cfg=act_cfg,
                            stride=stride,
                            dilation=dil,
                            dropout=drop,
                            dist_bn=self.dist_bn,
                        ),
                    )
                )

                # Update channels and p_keep
                in_channels = channels[mod_id][-1]

            # Create module
            if mod_id < 2:
                self.add_module(
                    "pool%d" % (mod_id + 2), nn.MaxPool2d(3, stride=2, padding=1)
                )
            self.add_module("mod%d" % (mod_id + 2), nn.Sequential(OrderedDict(blocks)))

        # Pooling and predictor
        # self.bn_out = bnrelu(in_channels, norm_cfg=norm_cfg, act_cfg=act_cfg)

    def forward(self, img):
        r1 = self.mod1(img)
        r2 = self.mod2(self.pool2(r1))
        r3 = self.mod3(self.pool3(r2))
        r4 = self.mod4(r3)
        r5 = self.mod5(r4)
        r6 = self.mod6(r5)
        r7 = self.mod7(r6)
        # out = self.bn_out(out)

        outs = [r1, r2, r3, r4, r5, r6, r7]
        outs = [outs[i] for i in self.out_indices]

        return outs



#!/usr/bin/env python3

"""Custom Joint Dataset Classes."""

import os.path as osp

import mmcv
from mmcv.utils import print_log
from pyEdgeEval.edge_tools import Mask2Edge, InstanceMask2Edge
from blette.datasets.pipelines import (
    Compose,
    EdgeFormatBundle,
    LoadAnnotations,
    LoadEdges,
    BinaryEdgeFormatBundle,
)

from .base_dataset import BaseDataset
from potato.datasets.pipelines import JointFormatBundle, JointBinaryFormatBundle
from potato.utils import get_root_logger


class CustomJointDataset(BaseDataset):
    """Custom dataset class where we use the preprocessed edges."""

    def __init__(
        self,
        edge_dir=None,
        edge_map_suffix="_edge.png",
        inst_sensitive=False,
        gt_seg_loader_cfg=None,
        gt_edge_loader_cfg=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.edge_dir = edge_dir
        self.edge_map_suffix = edge_map_suffix

        # FIXME: not being used
        self.inst_sensitive = inst_sensitive

        # TODO: create a custom joint loader and format
        self.gt_seg_loader = (
            LoadAnnotations()
            if gt_seg_loader_cfg is None
            else LoadAnnotations(**gt_seg_loader_cfg)
        )

        if gt_edge_loader_cfg is None:
            gt_edge_loader_cfg = dict(format=True)
        else:
            gt_edge_loader_cfg = dict(
                format=True,
                **gt_edge_loader_cfg,
            )
        self.gt_edge_loader = LoadEdges(**gt_edge_loader_cfg)

        # join paths if data_root is specified
        if self.data_root is not None:
            if not (self.edge_dir is None or osp.isabs(self.edge_dir)):
                self.edge_dir = osp.join(self.data_root, self.edge_dir)

        # load annotations
        self.img_infos = self.load_annotations(
            img_dir=self.img_dir,
            img_suffix=self.img_suffix,
            seg_map_suffix=self.seg_map_suffix,
            edge_map_suffix=self.edge_map_suffix,
            split=self.split,
        )

    def load_annotations(
        self,
        img_dir,
        img_suffix,
        seg_map_suffix,
        edge_map_suffix,
        split,
    ):
        img_infos = []
        if split is not None:
            with open(split) as f:
                for line in f:
                    img_name = line.strip()
                    img_info = dict(filename=img_name + img_suffix)
                    seg_map = img_name + seg_map_suffix
                    edge_map = img_name + edge_map_suffix
                    ann = dict(
                        seg_map=seg_map,
                        edge_map=edge_map,
                    )
                    img_info["ann"] = ann
                    img_infos.append(img_info)
        else:
            for img in mmcv.scandir(img_dir, img_suffix, recursive=True):
                img_info = dict(filename=img)
                seg_map = img.replace(img_suffix, seg_map_suffix)
                edge_map = img.replace(img_suffix, edge_map_suffix)
                ann = dict(
                    seg_map=seg_map,
                    edge_map=edge_map,
                )
                img_info["ann"] = ann
                img_infos.append(img_info)
            img_infos = sorted(img_infos, key=lambda x: x["filename"])

        print_log(f"Loaded {len(img_infos)} images", logger=get_root_logger())
        return img_infos

    def pre_pipeline(self, results):
        results["seg_fields"] = []
        results["img_prefix"] = self.img_dir
        results["seg_prefix"] = self.ann_dir
        results["edge_prefix"] = self.edge_dir
        # need for converting labelids to trainids
        results["num_classes"] = len(self.CLASSES)


class OTFCustomJointDataset(BaseDataset):
    """Custom dataset class where we generate edges OTF"""

    IDS = None
    inst_sensitive = None
    mask2edge = None

    def __init__(
        self,
        inst_dir=None,
        inst_map_suffix="_inst.png",
        inst_sensitive=False,
        gt_seg_loader_cfg=None,
        gt_edge_loader_cfg=None,
        class_names=None,
        labelIds=None,
        inst_labelIds=None,
        ignore_indices=[],
        label2trainId=None,
        radius=2,
        **kwargs,
    ):
        super().__init__(**kwargs)

        if class_names:
            if self.CLASSES is not None:
                print_log(
                    "Resetting cls.CLASSES, which might be unintentional",
                    logger=get_root_logger(),
                )
            self.CLASSES = class_names
            self.IDS = list(range(self.CLASSES))

        self.inst_dir = inst_dir
        self.inst_map_suffix = inst_map_suffix
        self.inst_sensitive = inst_sensitive

        # initialize mask2edge
        labelIds = labelIds if labelIds else self.IDS
        if self.inst_sensitive:
            assert (
                inst_labelIds
            ), "ERR: `inst_labelIds` is needed for instance sensitive OTF"
            self.mask2edge = InstanceMask2Edge(
                labelIds=labelIds,
                inst_labelIds=inst_labelIds,
                ignore_indices=ignore_indices,
                label2trainId=label2trainId,
                radius=radius,
                use_cv2=True,
                quality=0,
            )
        else:
            self.mask2edge = Mask2Edge(
                labelIds=labelIds,
                ignore_indices=ignore_indices,
                label2trainId=label2trainId,
                radius=radius,
                use_cv2=True,
                quality=0,
            )

        self.gt_seg_loader = Compose(
            [
                LoadAnnotations()
                if gt_seg_loader_cfg is None
                else LoadAnnotations(**gt_seg_loader_cfg),
                JointFormatBundle(),  # need to convert labels to trainIds
            ]
        )

        self.gt_edge_loader = Compose(
            [
                LoadAnnotations()
                if gt_edge_loader_cfg is None
                else LoadAnnotations(**gt_edge_loader_cfg),
                EdgeFormatBundle(),  # need to convert seg to edges
            ]
        )

        # join paths if data_root is specified
        if self.data_root is not None:
            if not (self.inst_dir is None or osp.isabs(self.inst_dir)):
                self.inst_dir = osp.join(self.data_root, self.inst_dir)

        # load annotations
        assert self.ann_dir is not None
        self.img_infos = self.load_annotations(
            img_dir=self.img_dir,
            img_suffix=self.img_suffix,
            seg_map_suffix=self.seg_map_suffix,
            inst_map_suffix=self.inst_map_suffix,
            inst_sensitive=self.inst_sensitive,
            split=self.split,
        )

    def load_annotations(
        self,
        img_dir,
        img_suffix,
        seg_map_suffix,
        inst_map_suffix,
        inst_sensitive,
        split,
    ):
        img_infos = []
        if split is not None:
            with open(split) as f:
                for line in f:
                    img_name = line.strip()
                    img_info = dict(filename=img_name + img_suffix)
                    seg_map = img_name + seg_map_suffix
                    if inst_sensitive:
                        # need instance map for instance aware segmentation (otf)
                        inst_map = img_name + inst_map_suffix
                        ann = dict(
                            seg_map=seg_map,
                            inst_map=inst_map,
                        )
                    else:
                        ann = dict(
                            seg_map=seg_map,
                        )
                    img_info["ann"] = ann
                    img_infos.append(img_info)
        else:
            for img in mmcv.scandir(img_dir, img_suffix, recursive=True):
                img_info = dict(filename=img)
                seg_map = img.replace(img_suffix, seg_map_suffix)
                if inst_sensitive:
                    inst_map = img.replace(img_suffix, inst_map_suffix)
                    ann = dict(
                        seg_map=seg_map,
                        inst_map=inst_map,
                    )
                else:
                    ann = dict(
                        seg_map=seg_map,
                    )
                img_info["ann"] = ann
                img_infos.append(img_info)
            img_infos = sorted(img_infos, key=lambda x: x["filename"])

        print_log(f"Loaded {len(img_infos)} images", logger=get_root_logger())
        return img_infos

    def pre_pipeline(self, results):
        results["seg_fields"] = []
        results["img_prefix"] = self.img_dir
        results["seg_prefix"] = self.ann_dir
        results["inst_prefix"] = self.inst_dir
        results["inst_sensitive"] = self.inst_sensitive
        results["mask2edge"] = self.mask2edge


class OTFCustomBinaryJointDataset(BaseDataset):

    IDS = None
    mask2edge = None

    def __init__(
        self,
        gt_seg_loader_cfg=None,
        gt_edge_loader_cfg=None,
        ignore_indices=[],
        labelIds=None,
        label2trainId=None,
        radius=1,
        selected_label=1,  # index for the edge
        **kwargs,
    ):
        super().__init__(**kwargs)

        labelIds = labelIds if labelIds else self.IDS
        self.mask2edge = Mask2Edge(
            labelIds=labelIds,
            ignore_indices=ignore_indices,
            label2trainId=label2trainId,
            radius=radius,
            use_cv2=True,
            quality=0,
        )

        self.gt_seg_loader = Compose(
            [
                LoadAnnotations()
                if gt_seg_loader_cfg is None
                else LoadAnnotations(**gt_seg_loader_cfg),
                JointBinaryFormatBundle(selected_label=selected_label),
            ]
        )

        assert selected_label is not None
        self.gt_edge_loader = Compose(
            [
                LoadAnnotations()
                if gt_edge_loader_cfg is None
                else LoadAnnotations(**gt_edge_loader_cfg),
                BinaryEdgeFormatBundle(
                    selected_label=selected_label,
                ),
            ]
        )

        # load annotations
        assert self.ann_dir is not None
        self.img_infos = self.load_annotations(
            img_dir=self.img_dir,
            img_suffix=self.img_suffix,
            seg_map_suffix=self.seg_map_suffix,
            split=self.split,
        )

    def load_annotations(
        self,
        img_dir,
        img_suffix,
        seg_map_suffix,
        split,
    ):
        img_infos = []
        if split is not None:
            with open(split) as f:
                for line in f:
                    img_name = line.strip()
                    img_info = dict(filename=img_name + img_suffix)
                    seg_map = img_name + seg_map_suffix
                    img_info["ann"] = dict(seg_map=seg_map)
                    img_infos.append(img_info)
        else:
            for img in mmcv.scandir(img_dir, img_suffix, recursive=True):
                img_info = dict(filename=img)
                seg_map = img.replace(img_suffix, seg_map_suffix)
                img_info["ann"] = dict(seg_map=seg_map)
                img_infos.append(img_info)
            img_infos = sorted(img_infos, key=lambda x: x["filename"])

        print_log(f"Loaded {len(img_infos)} images", logger=get_root_logger())
        return img_infos

    def pre_pipeline(self, results):
        results["seg_fields"] = []
        results["img_prefix"] = self.img_dir
        results["seg_prefix"] = self.ann_dir
        results["mask2edge"] = self.mask2edge



#!/usr/bin/env python3

from .builder import (
    DATASETS,
    PIPELINES,
    build_dataloader,
    build_dataset,
)
from .mmseg_dataset_wrapper import FBOUNDARY_EVAL_DATASET_MAPPING
from .pipelines import *  # noqa: F401,F403

from .custom import (
    CustomJointDataset,
    OTFCustomJointDataset,
    OTFCustomBinaryJointDataset,
)
from .cityscapes import JointCityscapesDataset, OTFJointCityscapesDataset
from .isprs import JointISPRSDataset, OTFJointISPRSDataset
from .potsdam import JointPotsdamDataset, OTFJointPotsdamDataset
from .loveda import JointLoveDADataset, OTFJointLoveDADataset
from .inria_aerial import INRIAAerialDataset, OTFINRIAAerialDataset

__all__ = [
    "DATASETS",
    "PIPELINES",
    "FBOUNDARY_EVAL_DATASET_MAPPING",
    "build_dataloader",
    "build_dataset",
    "CustomJointDataset",
    "OTFCustomJointDataset",
    "OTFCustomBinaryJointDataset",
    "JointCityscapesDataset",
    "OTFJointCityscapesDataset",
    "JointISPRSDataset",
    "OTFJointISPRSDataset",
    "JointPotsdamDataset",
    "OTFJointPotsdamDataset",
    "JointLoveDADataset",
    "OTFJointLoveDADataset",
    "INRIAAerialDataset",
    "OTFINRIAAerialDataset",
]



#!/usr/bin/env python3

import glob
import os.path as osp
import warnings


def find_latest_checkpoint(path, suffix="pth"):
    """This function is for finding the latest checkpoint.

    It will be used when automatically resume, modified from
    https://github.com/open-mmlab/mmdetection/blob/dev-v2.20.0/mmdet/utils/misc.py

    Args:
        path (str): The path to find checkpoints.
        suffix (str): File extension for the checkpoint. Defaults to pth.

    Returns:
        latest_path(str | None): File path of the latest checkpoint.
    """
    if not osp.exists(path):
        warnings.warn("The path of the checkpoints doesn't exist.")
        return None
    if osp.exists(osp.join(path, f"latest.{suffix}")):
        return osp.join(path, f"latest.{suffix}")

    checkpoints = glob.glob(osp.join(path, f"*.{suffix}"))
    if len(checkpoints) == 0:
        warnings.warn("The are no checkpoints in the path")
        return None
    latest = -1
    latest_path = ""
    for checkpoint in checkpoints:
        if len(checkpoint) < len(latest_path):
            continue
        # `count` is iteration number, as checkpoints are saved as
        # 'iter_xx.pth' or 'epoch_xx.pth' and xx is iteration number.
        count = int(osp.basename(checkpoint).split("_")[-1].split(".")[0])
        if count > latest:
            latest = count
            latest_path = checkpoint
    return latest_path



"""
Copyright (C) 2019 NVIDIA Corporation.  All rights reserved.
Licensed under the CC BY-NC-SA 4.0 license (https://creativecommons.org/licenses/by-nc-sa/4.0/legalcode).
# Code adapted from:
# https://github.com/fperazzi/davis/blob/master/python/lib/davis/measures/f_boundary.py
#
# Source License
#
# BSD 3-Clause License
#
# Copyright (c) 2017,
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.s
##############################################################################
#
# Based on:
# ----------------------------------------------------------------------------
# A Benchmark Dataset and Evaluation Methodology for Video Object Segmentation
# Copyright (c) 2016 Federico Perazzi
# Licensed under the BSD License [see LICENSE for details]
# Written by Federico Perazzi
# ----------------------------------------------------------------------------
"""


import numpy as np
from multiprocessing import Pool

# from tqdm import tqdm

""" Utilities for computing, reading and saving benchmark evaluation."""


def eval_mask_boundary_batch(
    seg_mask, gt_mask, num_classes, num_proc=16, bound_th=0.008
):
    """
    Compute F score for a segmentation mask
    Arguments:
        seg_mask (ndarray): segmentation mask prediction
        gt_mask (ndarray): segmentation mask ground truth
        num_classes (int): number of classes
    Returns:
        F (float): mean F score across all classes
        Fpc (listof float): F score per class
    """
    p = Pool(processes=num_proc)
    batch_size = seg_mask.shape[0]

    Fpc = np.zeros(num_classes)
    Fc = np.zeros(num_classes)
    for class_id in range(num_classes):
        args = [
            (
                (seg_mask[i] == class_id).astype(np.uint8),
                (gt_mask[i] == class_id).astype(np.uint8),
                gt_mask[i] == 255,
                bound_th,
            )
            for i in range(batch_size)
        ]
        temp = p.imap(db_eval_boundary_wrapper, args)

        Fs = []
        for t, _ in temp:
            Fs.append(t)

        Fs = np.array(Fs)
        # Fs = temp[:, 0]
        _valid = ~np.isnan(Fs)
        Fc[class_id] = np.sum(_valid)
        Fs[np.isnan(Fs)] = 0
        Fpc[class_id] = sum(Fs)

    p.close()
    p.join()
    return Fpc, Fc


def eval_mask_boundary(seg_mask, gt_mask, num_classes, num_proc=16, bound_th=0.008):
    """
    Compute F score for a segmentation mask
    Arguments:
        seg_mask (ndarray): segmentation mask prediction
        gt_mask (ndarray): segmentation mask ground truth
        num_classes (int): number of classes
    Returns:
        F (float): mean F score across all classes
        Fpc (listof float): F score per class
    """
    Fpc = np.zeros(num_classes)
    Fc = np.zeros(num_classes)
    for class_id in range(num_classes):

        # arguments
        foreground_mask = seg_mask == class_id
        cls_gt_mask = (gt_mask == class_id).astype(np.uint8)
        ignore = gt_mask == 255

        # only returns a single value
        F, _ = db_eval_boundary(foreground_mask, cls_gt_mask, ignore, bound_th)
        # print(class_id, F)
        # check if it is nan
        if not np.isnan(F):
            # counter
            Fc[class_id] += 1
            # perclass F score
            Fpc[class_id] = F

    return Fpc, Fc


def pre_eval_to_boundary(results):
    assert isinstance(results, list)

    # NOTE: should be list of tuples
    # convert to tuples of list
    Fpcs, Fcs = zip(*results)

    # N, Classes
    Fpcs = np.stack(Fpcs, axis=0)
    Fcs = np.stack(Fcs, axis=0)

    print("Fpcs:", Fpcs.shape)
    print("Fc:", Fcs.shape)

    Fpc = np.sum(Fpcs, axis=0)
    Fc = np.sum(Fcs, axis=0)

    # print(Fpc, Fc)

    return Fpc / Fc


# def db_eval_boundary_wrapper_wrapper(args):
#    seg_mask, gt_mask, class_id, batch_size, Fpc = args
#    print("class_id:" + str(class_id))
#    p = Pool(processes=10)
#    args = [((seg_mask[i] == class_id).astype(np.uint8),
#             (gt_mask[i] == class_id).astype(np.uint8))
#             for i in range(batch_size)]
#    Fs = p.map(db_eval_boundary_wrapper, args)
#    Fpc[class_id] = sum(Fs)
#    return


def db_eval_boundary_wrapper(args):
    foreground_mask, gt_mask, ignore, bound_th = args
    return db_eval_boundary(foreground_mask, gt_mask, ignore, bound_th)


def db_eval_boundary(foreground_mask, gt_mask, ignore_mask, bound_th=0.008):
    """
    Compute mean,recall and decay from per-frame evaluation.
    Calculates precision/recall for boundaries between foreground_mask and
    gt_mask using morphological operators to speed it up.
    Arguments:
            foreground_mask (ndarray): binary segmentation image.
            gt_mask         (ndarray): binary annotated image.
    Returns:
            F (float): boundaries F-measure
            P (float): boundaries precision
            R (float): boundaries recall
    """
    assert np.atleast_3d(foreground_mask).shape[2] == 1

    bound_pix = (
        bound_th
        if bound_th >= 1
        else np.ceil(bound_th * np.linalg.norm(foreground_mask.shape))
    )

    # print(bound_pix)
    # print(gt.shape)
    # print(np.unique(gt))
    foreground_mask[ignore_mask] = 0
    gt_mask[ignore_mask] = 0

    # Get the pixel boundaries of both masks
    fg_boundary = seg2bmap(foreground_mask)
    gt_boundary = seg2bmap(gt_mask)

    from skimage.morphology import binary_dilation, disk

    fg_dil = binary_dilation(fg_boundary, disk(bound_pix))
    gt_dil = binary_dilation(gt_boundary, disk(bound_pix))

    # Get the intersection
    gt_match = gt_boundary * fg_dil
    fg_match = fg_boundary * gt_dil

    # Area of the intersection
    n_fg = np.sum(fg_boundary)
    n_gt = np.sum(gt_boundary)

    # Compute precision and recall
    if n_fg == 0 and n_gt > 0:
        precision = 1
        recall = 0
    elif n_fg > 0 and n_gt == 0:
        precision = 0
        recall = 1
    elif n_fg == 0 and n_gt == 0:
        precision = 1
        recall = 1
    else:
        precision = np.sum(fg_match) / float(n_fg)
        recall = np.sum(gt_match) / float(n_gt)

    # Compute F measure
    if precision + recall == 0:
        F = 0
    else:
        # alpha is 0.5
        F = 2 * precision * recall / (precision + recall)

    return F, precision


def seg2bmap(seg, width=None, height=None):
    """
    From a segmentation, compute a binary boundary map with 1 pixel wide
    boundaries.  The boundary pixels are offset by 1/2 pixel towards the
    origin from the actual segment boundary.
    Arguments:
            seg     : Segments labeled from 1..k.
            width	  :	Width of desired bmap  <= seg.shape[1]
            height  :	Height of desired bmap <= seg.shape[0]
    Returns:
            bmap (ndarray):	Binary boundary map.
     David Martin <dmartin@eecs.berkeley.edu>
     January 2003
    """

    seg = seg.astype(np.bool)
    seg[seg > 0] = 1

    assert np.atleast_3d(seg).shape[2] == 1

    width = seg.shape[1] if width is None else width
    height = seg.shape[0] if height is None else height

    h, w = seg.shape[:2]

    ar1 = float(width) / float(height)
    ar2 = float(w) / float(h)

    assert not (
        width > w | height > h | abs(ar1 - ar2) > 0.01
    ), "Can" "t convert %dx%d seg to %dx%d bmap." % (w, h, width, height)

    e = np.zeros_like(seg)
    s = np.zeros_like(seg)
    se = np.zeros_like(seg)

    e[:, :-1] = seg[:, 1:]
    s[:-1, :] = seg[1:, :]
    se[:-1, :-1] = seg[1:, 1:]

    b = seg ^ e | seg ^ s | seg ^ se
    b[-1, :] = seg[-1, :] ^ e[-1, :]
    b[:, -1] = seg[:, -1] ^ s[:, -1]
    b[-1, -1] = 0

    if w == width and h == height:
        bmap = b
    else:
        bmap = np.zeros((height, width))
        for x in range(w):
            for y in range(h):
                if b[y, x]:
                    j = 1 + np.floor((y - 1) + height / h)
                    i = 1 + np.floor((x - 1) + width / h)
                    bmap[j, i] = 1

    return bmap
